# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_log import log
import yaml

LOG = log.getLogger(__name__)


def list_files(dir_path,
               suffix=None,
               with_pathname=False,
               with_exception=False):
    if not os.path.isdir(dir_path):
        return []

    try:
        file_list = os.listdir(dir_path)
    except Exception as e:
        if with_exception:
            raise
        else:
            LOG.error("Fails to list files in %s: %s" % (dir_path, e))
            return []
    if suffix:
        file_list = [os.path.join(dir_path, f) if with_pathname else f
                     for f in file_list if f.endswith(suffix)]

    return file_list


def load_yaml_files(dir_path, with_exception=False):
    files = list_files(dir_path, '.yaml', with_pathname=True)

    yaml_files = []
    for f in files:
        config = load_yaml_file(f, with_exception)
        if config:
            yaml_files.append(config)

    return yaml_files


def load_yaml_file(full_path, with_exception=False):
    if not os.path.isfile(full_path):
        LOG.error("File doesn't exist: %s." % full_path)
        return None

    with open(full_path, 'r') as stream:
        try:
            return yaml.load(stream, Loader=yaml.BaseLoader)
        except Exception as e:
            if with_exception:
                raise
            else:
                LOG.error("Fails to parse file: %s. %s" % (full_path, e))
                return None



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import re

from oslo_log import log
from voluptuous import All
from voluptuous import Any
from voluptuous import Error
from voluptuous import Invalid
from voluptuous import Optional
from voluptuous import Required
from voluptuous import Schema

from vitrage.common.constants import EntityCategory
from vitrage.evaluator.actions.base import action_types
from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_validation.base import get_correct_result
from vitrage.evaluator.template_validation.base import get_fault_result
from vitrage.evaluator.template_validation.base import get_status_code
from vitrage.evaluator.template_validation.status_messages import status_msgs

LOG = log.getLogger(__name__)

RESULT_DESCRIPTION = 'Template syntax validation'
EXCEPTION = 'exception'


def syntax_validation(template_conf):
    if template_conf.get(EXCEPTION):
        result = get_fault_result(RESULT_DESCRIPTION,
                                  5,
                                  msg=status_msgs[5] +
                                  template_conf.get(EXCEPTION))
    else:
        result = _validate_template_sections(template_conf)

        if result.is_valid_config:
            metadata = template_conf[TemplateFields.METADATA]
            result = _validate_metadata_section(metadata)

        if result.is_valid_config and TemplateFields.INCLUDES in template_conf:
            includes = template_conf[TemplateFields.INCLUDES]
            result = _validate_includes_section(includes)

        if result.is_valid_config and \
           TemplateFields.DEFINITIONS in template_conf:
            definitions = template_conf[TemplateFields.DEFINITIONS]
            has_includes = TemplateFields.INCLUDES in template_conf
            result = _validate_definitions_section(definitions, has_includes)

        if result.is_valid_config:
            scenarios = template_conf[TemplateFields.SCENARIOS]
            result = _validate_scenarios_section(scenarios)

    return result


def def_template_syntax_validation(def_template_conf):
    result = _validate_def_template_template_sections(def_template_conf)

    if TemplateFields.INCLUDES in def_template_conf or \
       TemplateFields.SCENARIOS in def_template_conf:
        LOG.error('%s status code: %s' % (status_msgs[143], 143))
        return get_fault_result(RESULT_DESCRIPTION, 143)

    if result.is_valid_config:
        metadata = def_template_conf[TemplateFields.METADATA]
        result = _validate_metadata_section(metadata)

    if result.is_valid_config:
        definitions = def_template_conf[TemplateFields.DEFINITIONS]
        result = _validate_definitions_section(definitions, False)

    return result


def _validate_def_template_template_sections(def_template_conf):
    schema = Schema({
        Required(TemplateFields.DEFINITIONS, msg=21): dict,
        Required(TemplateFields.METADATA, msg=62): dict,
    })
    return _validate_dict_schema(schema, def_template_conf)


def _validate_template_sections(template_conf):
    any_str = Any(str)
    paramsSchema = Schema({
        any_str: Any(any_str, Schema({
            Optional(TemplateFields.DESCRIPTION): any_str,
            Optional(TemplateFields.DEFAULT): any_str,
        })),
    })

    if TemplateFields.INCLUDES in template_conf:
        schema = Schema({
            Optional(TemplateFields.DEFINITIONS): dict,
            Required(TemplateFields.METADATA, msg=62): dict,
            Required(TemplateFields.SCENARIOS, msg=80): list,
            Optional(TemplateFields.INCLUDES): list,
            Optional(TemplateFields.PARAMETERS): paramsSchema,
        })
    else:
        schema = Schema({
            Required(TemplateFields.DEFINITIONS, msg=21): dict,
            Required(TemplateFields.METADATA, msg=62): dict,
            Required(TemplateFields.SCENARIOS, msg=80): list,
            Optional(TemplateFields.INCLUDES): list,
            Optional(TemplateFields.PARAMETERS): paramsSchema,
        })
    return _validate_dict_schema(schema, template_conf)


def _validate_metadata_section(metadata):
    any_str = Any(str)

    schema = Schema({
        TemplateFields.VERSION: any_str,
        Required(TemplateFields.NAME, msg=60): any_str,
        TemplateFields.DESCRIPTION: any_str,
        TemplateFields.TYPE: any_str,
    })
    return _validate_dict_schema(schema, metadata)


def _validate_includes_section(includes):
    any_str = Any(str)
    if not includes:
        LOG.error('%s status code: %s' % (status_msgs[140], 140))
        return get_fault_result(RESULT_DESCRIPTION, 140)

    for name in includes:
        schema = Schema({
            Required(TemplateFields.NAME, msg=141): any_str
        })
        result = _validate_name_schema(schema, name)
        if not result.is_valid_config:
            return result

    return result


def _validate_name_schema(schema, name):
    try:
        schema(name)
    except Error as e:

        status_code = get_status_code(e)
        if status_code:
            msg = status_msgs[status_code]
        else:
            # General syntax error
            status_code = 4
            msg = status_msgs[4] + str(e)

        LOG.error('%s status code: %s' % (msg, status_code))
        return get_fault_result(RESULT_DESCRIPTION, status_code, msg)

    return get_correct_result(RESULT_DESCRIPTION)


def _validate_definitions_section(definitions, has_includes):
    # Entities are required if there are no relationships, or if there are
    # relationships and no imported entities from a definition template
    # (otherwise the template is empty)
    if TemplateFields.RELATIONSHIPS not in definitions \
            or (definitions[TemplateFields.RELATIONSHIPS]
                and not has_includes):
        schema = Schema({
            Required(TemplateFields.ENTITIES, msg=20): list,
            TemplateFields.RELATIONSHIPS: list
        })
        result = _validate_dict_schema(schema, definitions)

    else:
        result = get_correct_result(RESULT_DESCRIPTION)

    if result.is_valid_config and TemplateFields.ENTITIES in definitions:
        entities = definitions[TemplateFields.ENTITIES]
        result = _validate_entities(entities, has_includes)

        relationships = definitions.get(TemplateFields.RELATIONSHIPS, None)
        if result.is_valid_config and relationships:
            return _validate_relationships(relationships)

    return result


def _validate_entities(entities, has_includes):
    if not entities and not has_includes:
        LOG.error('%s status code: %s' % (status_msgs[43], 43))
        return get_fault_result(RESULT_DESCRIPTION, 43)

    for entity in entities:

        schema = Schema({
            Required(TemplateFields.ENTITY, msg=46): dict,
        })
        result = _validate_dict_schema(schema, entity)

        if result.is_valid_config:
            result = _validate_entity_dict(entity[TemplateFields.ENTITY])

        if not result.is_valid_config:
            return result

    return result


def _validate_entity_dict(entity_dict):
    any_str = Any(str)
    schema = Schema({
        Required(TemplateFields.CATEGORY, msg=42):
            All(_validate_category_field()),
        TemplateFields.TYPE: any_str,
        Required(TemplateFields.TEMPLATE_ID, msg=41):
            All(_validate_template_id_value())
    }, extra=True)

    return _validate_dict_schema(schema, entity_dict)


def _validate_relationships(relationships):
    for relationship in relationships:

        schema = Schema({
            Required(TemplateFields.RELATIONSHIP, msg=101): dict,
        })
        result = _validate_dict_schema(schema, relationship)

        if result.is_valid_config:
            relationship_dict = relationship[TemplateFields.RELATIONSHIP]
            result = _validate_relationship_dict(relationship_dict)

            if not result.is_valid_config:
                return result
    return result


def _validate_relationship_dict(relationship_dict):
    any_str = Any(str)
    schema = Schema({
        Required(TemplateFields.SOURCE, msg=102): any_str,
        Required(TemplateFields.TARGET, msg=103): any_str,
        Required(TemplateFields.RELATIONSHIP_TYPE, msg=100): any_str,
        Required(TemplateFields.TEMPLATE_ID, msg=104):
            All(_validate_template_id_value())
    })
    return _validate_dict_schema(schema, relationship_dict)


def _validate_scenarios_section(scenarios):
    if not scenarios:
        LOG.error('%s status code: %s' % (status_msgs[81], 81))
        return get_fault_result(RESULT_DESCRIPTION, 81)

    for scenario in scenarios:

        schema = Schema({
            Required(TemplateFields.SCENARIO, msg=82): dict,
        })
        result = _validate_dict_schema(schema, scenario)

        if result.is_valid_config:
            result = _validate_scenario(scenario[TemplateFields.SCENARIO])

            if not result.is_valid_config:
                return result

    return result


def _validate_scenario(scenario):
    any_str = Any(str)
    schema = Schema({
        Required(TemplateFields.CONDITION, msg=83): any_str,
        Required(TemplateFields.ACTIONS, msg=84): list
    })
    result = _validate_dict_schema(schema, scenario)

    if result.is_valid_config:
        return _validate_actions_schema(scenario[TemplateFields.ACTIONS])

    return result


def _validate_actions_schema(actions):
    if not actions:
        LOG.error('%s status code: %s' % (status_msgs[121], 121))
        return get_fault_result(RESULT_DESCRIPTION, 121)

    for action in actions:

        schema = Schema({
            Required(TemplateFields.ACTION, msg=122): dict,
        })
        result = _validate_dict_schema(schema, action)

        if result.is_valid_config:
            result = _validate_action_schema(action[TemplateFields.ACTION])

        if not result.is_valid_config:
            return result

    return result


def _validate_action_schema(action):
    schema = Schema({
        Required(TemplateFields.ACTION_TYPE, msg=123):
            _validate_action_type_field(),
        TemplateFields.PROPERTIES: dict,
        TemplateFields.ACTION_TARGET: dict
    })
    return _validate_dict_schema(schema, action)


def _validate_dict_schema(schema, value):
    try:
        schema(value)
    except Error as e:
        status_code = get_status_code(e)
        msg = status_msgs[status_code] + str(e)
        LOG.error('%s status code: %s' % (msg, status_code))
        return get_fault_result(RESULT_DESCRIPTION, status_code, msg)

    return get_correct_result(RESULT_DESCRIPTION)


def _validate_template_id_value(msg=None):
    def f(v):
        if re.match("_*[a-zA-Z]+\\w*", str(v)):
            return str(v)
        else:
            raise Invalid(msg or 1)

    return f


def _validate_category_field(msg=None):
    def f(v):
        if str(v) in EntityCategory.categories():
            return str(v)
        else:
            raise Invalid(msg or 45)

    return f


def _validate_action_type_field(msg=None):
    def f(v):
        if str(v) in action_types:
            return str(v)
        else:
            raise Invalid(msg or 120)

    return f



#    Copyright 2016 - Nokia
#    Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from testtools import matchers

from vitrage.hacking import checks
from vitrage.tests import base
from vitrage.tests.base import IsEmpty


class HackingTestCase(base.BaseTest):
    def test_assert_true_instance(self):
        self.assertThat(list(checks.assert_true_instance(
            "self.assertTrue(isinstance(e, "
            "exception.BuildAbortException))")), matchers.HasLength(1))

        self.assertThat(list(checks.assert_true_instance("self.assertTrue("
                                                         ")")), IsEmpty())

    def test_assert_equal_type(self):
        self.assertThat(list(checks.assert_equal_type(
            "self.assertEqual(type(als['QuicAssist']), list)")),
            matchers.HasLength(1))

        self.assertThat(list(checks.assert_equal_type("self.assertTrue()")),
                        IsEmpty())

    def test_no_translate_logs(self):
        for log in checks._all_log_levels:
            bad = 'LOG.%s(_("Bad"))' % log
            self.assertThat(list(checks.no_translate_logs(bad)),
                            matchers.HasLength(1))
            # Catch abuses when used with a variable and not a literal
            bad = 'LOG.%s(_(msg))' % log
            self.assertThat(list(checks.no_translate_logs(bad)),
                            matchers.HasLength(1))

    def test_no_contextlib_nested(self):
        self.assertThat(list(checks.check_no_contextlib_nested(
            "with contextlib.nested(")), matchers.HasLength(1))

        self.assertThat(list(checks.check_no_contextlib_nested(
            "with nested(")), matchers.HasLength(1))

        self.assertThat(list(checks.check_no_contextlib_nested(
            "with foo as bar")), IsEmpty())

    def test_dict_constructor_with_list_copy(self):
        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "    dict([(i, connect_info[i])")), matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "    attrs = dict([(k, _from_json(v))")), matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "        type_names = dict((value, key) for key, value in")),
            matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "   dict((value, key) for key, value in")), matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "foo(param=dict((k, v) for k, v in bar.items()))")),
            matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            " dict([[i,i] for i in range(3)])")), matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "  dd = dict([i,i] for i in range(3))")), matchers.HasLength(1))

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "        create_kwargs = dict(snapshot=snapshot,")), IsEmpty())

        self.assertThat(list(checks.dict_constructor_with_list_copy(
            "      self._render_dict(xml, data_el, data.__dict__)")),
            IsEmpty())

    def test_check_python3_xrange(self):
        func = checks.check_python3_xrange
        self.assertThat(list(func('for i in xrange(10)')),
                        matchers.HasLength(1))
        self.assertThat(list(func('for i in range(10)')), IsEmpty())
        self.assertThat(list(func('for i in six.moves.range(10)')), IsEmpty())
        self.assertThat(list(func('testxrange(10)')), IsEmpty())

    def test_dict_iteritems(self):
        self.assertThat(list(checks.check_python3_no_iteritems(
            "obj.iteritems()")), matchers.HasLength(1))

        self.assertThat(list(checks.check_python3_no_iteritems(
            "six.iteritems(obj)")), IsEmpty())

        self.assertThat(list(checks.check_python3_no_iteritems(
            "obj.items()")), IsEmpty())

    def test_dict_iterkeys(self):
        self.assertThat(list(checks.check_python3_no_iterkeys(
            "obj.iterkeys()")), matchers.HasLength(1))

        self.assertThat(list(checks.check_python3_no_iterkeys(
            "six.iterkeys(obj)")), IsEmpty())

        self.assertThat(list(checks.check_python3_no_iterkeys(
            "obj.keys()")), IsEmpty())

    def test_dict_itervalues(self):
        self.assertThat(list(checks.check_python3_no_itervalues(
            "obj.itervalues()")), matchers.HasLength(1))

        self.assertThat(list(checks.check_python3_no_itervalues(
            "six.itervalues(ob)")), IsEmpty())

        self.assertThat(list(checks.check_python3_no_itervalues(
            "obj.values()")), IsEmpty())

    def test_no_mutable_default_args(self):
        self.assertThat(list(checks.no_mutable_default_args(
            " def fake_suds_context(calls={}):")), matchers.HasLength(1))

        self.assertThat(list(checks.no_mutable_default_args(
            "def get_info_from_bdm(virt_type, bdm, mapping=[])")),
            matchers.HasLength(1))

        self.assertThat(list(checks.no_mutable_default_args(
            "defined = []")), IsEmpty())

        self.assertThat(list(checks.no_mutable_default_args(
            "defined, undefined = [], {}")), IsEmpty())

    def test_no_log_warn(self):
        self.assertThat(list(checks.no_log_warn('LOG.warning("bl")')),
                        IsEmpty())
        self.assertThat(list(checks.no_log_warn('LOG.warn("foo")')),
                        matchers.HasLength(1))

    def test_asserttruefalse(self):
        true_fail_code1 = """
               test_bool = True
               self.assertEqual(True, test_bool)
               """
        true_fail_code2 = """
               test_bool = True
               self.assertEqual(test_bool, True)
               """
        true_pass_code = """
               test_bool = True
               self.assertTrue(test_bool)
               """
        false_fail_code1 = """
               test_bool = False
               self.assertEqual(False, test_bool)
               """
        false_fail_code2 = """
               test_bool = False
               self.assertEqual(test_bool, False)
               """
        false_pass_code = """
               test_bool = False
               self.assertFalse(test_bool)
               """
        self.assertThat(list(checks.check_assert_true_false(
            true_fail_code1)), matchers.HasLength(1))
        self.assertThat(list(checks.check_assert_true_false(
            true_fail_code2)), matchers.HasLength(1))
        self.assertThat(list(checks.check_assert_true_false(
            true_pass_code)), IsEmpty())
        self.assertThat(list(checks.check_assert_true_false(
            false_fail_code1)), matchers.HasLength(1))
        self.assertThat(list(checks.check_assert_true_false(
            false_fail_code2)), matchers.HasLength(1))
        self.assertFalse(list(checks.check_assert_true_false(false_pass_code)))



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from vitrage.entity_graph.mappings.operational_alarm_severity import \
    OperationalAlarmSeverity

from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps


false_ = 'False'
name_ = 'VM network problem'

GENERAL_OID = '1.3.6.1.4.1'
COMPANY_OID = '1.1.1'
ALARM_OBJECTS_OID = '1'
ALARM_PREFIX_OID = '2'
NAME_OID = '1'
IS_DELETED_OID = '2'
SEVERITY_OID = '3'

ALERT_OID = '.100000'

alarm_data = {VProps.VITRAGE_CATEGORY: EntityCategory.ALARM,
              VProps.NAME: name_,
              VProps.RESOURCE + '_' + VProps.VITRAGE_IS_DELETED: false_,
              VProps.RESOURCE + '_' + VProps.VITRAGE_IS_PLACEHOLDER: false_,
              VProps.VITRAGE_IS_DELETED: false_,
              VProps.VITRAGE_OPERATIONAL_SEVERITY:
                  OperationalAlarmSeverity.CRITICAL,
              VProps.RESOURCE:
                  {VProps.VITRAGE_IS_PLACEHOLDER: false_,
                   VProps.VITRAGE_IS_DELETED: false_}}

alert_details = {'oid': ALERT_OID, 'alarm_name': 'vitrageDeducedTest'}



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from vitrage.datasources.collectd.properties \
    import CollectdProperties as CProps

COLLECTD_HOST_PARAM = '${collectd_host}'


class CollectdMapper(object):

    def __init__(self, mapping):
        self.mapping = mapping

    def find(self, collectd_name):
        resource_value = self.mapping.get(collectd_name)
        if resource_value:
            return \
                {
                    CProps.RESOURCE_TYPE: resource_value[CProps.RESOURCE_TYPE],
                    CProps.RESOURCE_NAME: resource_value[CProps.RESOURCE_NAME]
                }
        return self.find_regex(collectd_name)

    def find_regex(self, collectd_name):
        for pattern, value in self.mapping.items():
            if re.match(pattern, collectd_name):
                type_, name = \
                    value[CProps.RESOURCE_TYPE], value[CProps.RESOURCE_NAME]
                if name == COLLECTD_HOST_PARAM:
                    return \
                        {
                            CProps.RESOURCE_TYPE: type_,
                            CProps.RESOURCE_NAME: collectd_name
                        }
                else:
                    return \
                        {
                            CProps.RESOURCE_TYPE: type_,
                            CProps.RESOURCE_NAME: name
                        }
        raise KeyError(collectd_name)



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_utils import importutils as utils

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import UpdateMethod
from vitrage.utils import opt_exists

CONF = cfg.CONF

drivers = {}


# noinspection PyProtectedMember
def get_drivers_by_name(driver_names):
    for d_name in driver_names:
        if not drivers.get(d_name):
            drivers[d_name] = utils.import_object(CONF[d_name].driver)
            drivers[d_name].__class__._datasource_name = d_name
    return [drivers[d_name] for d_name in driver_names]


def get_pull_drivers_names():
    return [name for name in CONF.datasources.types
            if CONF[name].update_method.lower() == UpdateMethod.PULL
            and opt_exists(CONF[name], DSOpts.CHANGES_INTERVAL)]


def get_push_drivers_names():
    return [name for name in CONF.datasources.types
            if CONF[name].update_method.lower() == UpdateMethod.PUSH]


def get_driver_class(driver_name):
    return utils.import_class(CONF[driver_name].driver)



# Copyright 2019 - China Mobile (SuZhou) Software Technology Co.,Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class StackProperties(object):
    STACK_NAME = 'stack_name'
    STACK_STATUS = 'stack_status'
    ID = 'id'
    STATE = 'state'
    CREATED_AT = 'created_at'
    STACK_IDENTITY = 'stack_identity'
    CREATION_TIME = 'creation_time'
    PROJECT = 'project'
    RESOURCES = 'resources'
    RESOURCES_TYPE = 'resource_type'
    PHYSICAL_RESOURCE_ID = 'physical_resource_id'
    TENANT_ID = 'tenant_id'



# Copyright 2020 - Inspur - Qitao
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.cetus.cetus_driver_base import CetusDriverBase
from vitrage.datasources.cetus.pod import CETUS_POD_DATASOURCE


class PodDriver(CetusDriverBase):

    def get_all(self, datasource_action):
        return self.make_pickleable(self._prepare_entities(
            self.list_all()),
            CETUS_POD_DATASOURCE,
            datasource_action)

    @staticmethod
    def _prepare_entities(pods):
        pods_dict = {}
        for pod in pods:

            p = {
                "name": pod["metadata"]["name"],
                "id": pod["metadata"]["uid"],
                "status": pod["status"]["phase"],
                "node": pod["metadata"]["labels"]["instance_id"]
            }

            p_id = pod["metadata"]["uid"]
            pods_dict[p_id] = p
        return [p for p in pods_dict.values()]



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import datetime

from oslo_config import cfg
from oslo_log import log as logging
from testtools import matchers

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import UpdateMethod
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.host.transformer import HostTransformer
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.nova.instance.transformer import InstanceTransformer
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver as mock_sync

LOG = logging.getLogger(__name__)


# noinspection PyProtectedMember
class BaseNovaInstanceTransformerTest(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    # noinspection PyAttributeOutsideInit,PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(BaseNovaInstanceTransformerTest, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=NOVA_INSTANCE_DATASOURCE)
        cls.conf.register_opts(cls._get_default_group_opts())
        cls.transformers[NOVA_HOST_DATASOURCE] = HostTransformer(
            cls.transformers)
        cls.transformers[NOVA_INSTANCE_DATASOURCE] = \
            InstanceTransformer(cls.transformers)

    def _test_create_placeholder_vertex(self):
        LOG.debug('Test create placeholder vertex')

        # Tests setup
        instance_id = 'Instance123'
        timestamp = datetime.datetime.utcnow()
        properties = {
            VProps.ID: instance_id,
            VProps.VITRAGE_TYPE: NOVA_INSTANCE_DATASOURCE,
            VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: timestamp
        }
        transformer = self.transformers[NOVA_INSTANCE_DATASOURCE]

        # Test action
        placeholder = \
            transformer.create_neighbor_placeholder_vertex(**properties)

        # Test assertions
        observed_uuid = placeholder.vertex_id
        expected_key = tbase.build_key(transformer._key_values(
            NOVA_INSTANCE_DATASOURCE,
            instance_id))
        expected_uuid = \
            TransformerBase.uuid_from_deprecated_vitrage_id(expected_key)
        self.assertEqual(expected_uuid, observed_uuid)

        observed_time = placeholder.get(VProps.VITRAGE_SAMPLE_TIMESTAMP)
        self.assertEqual(timestamp, observed_time)

        observed_type = placeholder.get(VProps.VITRAGE_TYPE)
        self.assertEqual(NOVA_INSTANCE_DATASOURCE, observed_type)

        observed_entity_id = placeholder.get(VProps.ID)
        self.assertEqual(instance_id, observed_entity_id)

        observed_vitrage_category = placeholder.get(VProps.VITRAGE_CATEGORY)
        self.assertEqual(EntityCategory.RESOURCE, observed_vitrage_category)

        vitrage_is_placeholder = placeholder.get(VProps.VITRAGE_IS_PLACEHOLDER)
        self.assertTrue(vitrage_is_placeholder)

    def _test_update_event_transform(self, instance_events):
        for event in instance_events:
            # Test action
            transformer = self.transformers[NOVA_INSTANCE_DATASOURCE]
            wrapper = transformer.transform(event)

            # Test assertions
            self._validate_update_vertex_props(
                transformer, wrapper.vertex, event)

            # Validate the neighbors: only one  valid host neighbor
            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_host_neighbor(neighbors[0], event)

            event_type = event[DSProps.EVENT_TYPE]
            if event_type == 'compute.instance.delete.end':
                self.assertEqual(GraphAction.DELETE_ENTITY, wrapper.action)
            elif event_type == 'compute.instance.create.start':
                self.assertEqual(GraphAction.CREATE_ENTITY, wrapper.action)
            else:
                self.assertEqual(GraphAction.UPDATE_ENTITY, wrapper.action)

    def _validate_snapshot_vertex_props(self, transformer, vertex, event):
        self.assertThat(vertex.properties, matchers.HasLength(15))
        self._validate_vertex_props(transformer, vertex, event)

    def _validate_update_vertex_props(self, transformer, vertex, event):
        self.assertThat(vertex.properties, matchers.HasLength(14))
        self._validate_vertex_props(transformer, vertex, event)

    def _validate_vertex_props(self, transformer, vertex, event):
        field_extractor = transformer._get_field_extractor(event)
        expected_id = field_extractor.entity_id(event)
        observed_id = vertex[VProps.ID]
        self.assertEqual(expected_id, observed_id)

        self.assertEqual(
            EntityCategory.RESOURCE,
            vertex[VProps.VITRAGE_CATEGORY]
        )
        self.assertEqual(NOVA_INSTANCE_DATASOURCE,
                         vertex[VProps.VITRAGE_TYPE])

        expected_project = field_extractor.tenant_id(event)
        observed_project = vertex[VProps.PROJECT_ID]
        self.assertEqual(expected_project, observed_project)

        expected_state = field_extractor.state(event)
        observed_state = vertex[VProps.STATE]
        self.assertEqual(expected_state, observed_state)

        expected_timestamp = event[DSProps.SAMPLE_DATE]
        observed_timestamp = vertex[VProps.VITRAGE_SAMPLE_TIMESTAMP]
        self.assertEqual(expected_timestamp, observed_timestamp)

        expected_name = field_extractor.name(event)
        observed_name = vertex[VProps.NAME]
        self.assertEqual(expected_name, observed_name)

        vitrage_is_placeholder = vertex[VProps.VITRAGE_IS_PLACEHOLDER]
        self.assertFalse(vitrage_is_placeholder)

        vitrage_is_deleted = vertex[VProps.VITRAGE_IS_DELETED]
        self.assertFalse(vitrage_is_deleted)

    def _validate_host_neighbor(self, h_neighbor, event):

        inst_transformer = self.transformers[NOVA_INSTANCE_DATASOURCE]
        field_extractor = inst_transformer._get_field_extractor(event)

        host_name = field_extractor.host(event)
        time = event[DSProps.SAMPLE_DATE]

        host_transformer = self.transformers[NOVA_HOST_DATASOURCE]
        properties = {
            VProps.ID: host_name,
            VProps.VITRAGE_TYPE: NOVA_HOST_DATASOURCE,
            VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: time
        }
        expected_neighbor = \
            host_transformer.create_neighbor_placeholder_vertex(**properties)
        self.assertEqual(expected_neighbor, h_neighbor.vertex)

        # Validate neighbor edge
        edge = h_neighbor.edge
        entity_key = inst_transformer._create_entity_key(event)
        entity_uuid = \
            TransformerBase.uuid_from_deprecated_vitrage_id(entity_key)
        self.assertEqual(edge.source_id, h_neighbor.vertex.vertex_id)
        self.assertEqual(edge.target_id, entity_uuid)
        self.assertEqual(edge.label, EdgeLabel.CONTAINS)

    def _test_create_entity_key(self):
        LOG.debug('Test get key from nova instance transformer')

        # Test setup
        spec_list = mock_sync.simple_instance_generators(
            host_num=1, vm_num=1, snapshot_events=1)
        instance_events = mock_sync.generate_random_events_list(spec_list)

        instance_transformer = self.transformers[NOVA_INSTANCE_DATASOURCE]
        for event in instance_events:
            # Test action
            observed_key = instance_transformer._create_entity_key(event)

            # Test assertions
            observed_key_fields = observed_key.split(
                TransformerBase.KEY_SEPARATOR)

            self.assertEqual(EntityCategory.RESOURCE, observed_key_fields[0])
            self.assertEqual(
                NOVA_INSTANCE_DATASOURCE,
                observed_key_fields[1]
            )

            instance_id = tbase.extract_field_value(event, 'id')

            self.assertEqual(instance_id, observed_key_fields[2])

            key_values = instance_transformer._key_values(
                NOVA_INSTANCE_DATASOURCE,
                instance_id)
            expected_key = tbase.build_key(key_values)

            self.assertEqual(expected_key, observed_key)

    def _test_build_instance_key(self):
        LOG.debug('Test build instance key')

        # Test setup
        instance_id = '456'
        expected_key = 'RESOURCE:nova.instance:%s' % instance_id

        instance_transformer = self.transformers[NOVA_INSTANCE_DATASOURCE]
        # Test action
        key_fields = instance_transformer._key_values(
            NOVA_INSTANCE_DATASOURCE,
            instance_id)

        # Test assertions
        observed_key = tbase.build_key(key_fields)
        self.assertEqual(expected_key, observed_key)

    @classmethod
    @abc.abstractmethod
    def _get_default_group_opts(cls):
        pass



# Copyright 2018 Samsung Electronics
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.resource_transformer_base import \
    ResourceTransformerBase
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.trove.instance import TROVE_INSTANCE_DATASOURCE
from vitrage.datasources.trove.properties import \
    TroveInstanceProperties as TProps
import vitrage.graph.utils as graph_utils


class TroveInstanceTransformer(ResourceTransformerBase):

    def _create_snapshot_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_vertex(self, entity_event):
        entity_id = entity_event[TProps.ID]
        name = entity_event[TProps.NAME]
        state = entity_event[TProps.STATE]
        project_id = entity_event[TProps.PROJECT_ID]
        update_timestamp = entity_event[TProps.UPDATE_TIMESTAMP]
        sample_timestamp = entity_event[DSProps.SAMPLE_DATE]
        metadata = {
            VProps.NAME: name,
            VProps.PROJECT_ID: project_id
        }
        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.RESOURCE,
            vitrage_type=TROVE_INSTANCE_DATASOURCE,
            vitrage_sample_timestamp=sample_timestamp,
            entity_id=entity_id,
            update_timestamp=update_timestamp,
            entity_state=state,
            metadata=metadata)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_entity_neighbours(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_entity_neighbours(entity_event)

    def _create_entity_neighbours(self, entity_event):
        server_neighbour = self._create_neighbor(
            entity_event,
            entity_event[TProps.SERVER_ID],
            NOVA_INSTANCE_DATASOURCE,
            EdgeLabel.CONTAINS,
            is_entity_source=True)
        return [server_neighbour]

    def _create_entity_key(self, entity_event):
        entity_id = entity_event[TProps.ID]
        key_fields = self._key_values(TROVE_INSTANCE_DATASOURCE, entity_id)
        return tbase.build_key(key_fields)

    @staticmethod
    def get_vitrage_type():
        return TROVE_INSTANCE_DATASOURCE



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json

from oslo_log import log

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.datasources.alarm_driver_base import AlarmDriverBase
from vitrage.datasources.aodh import AODH_DATASOURCE
from vitrage.datasources.aodh.properties import AodhEventType
from vitrage.datasources.aodh.properties import AodhExtendedAlarmType
from vitrage.datasources.aodh.properties import AodhProperties as AodhProps
from vitrage.datasources.aodh.properties import AodhState
from vitrage import os_clients
from vitrage.utils import datetime as datetime_utils

LOG = log.getLogger(__name__)


class AodhDriver(AlarmDriverBase):

    def __init__(self):
        super(AodhDriver, self).__init__()
        self._client = None
        self._init_aodh_event_actions()
        self._init_convert_aodh_alarm_rule_actions()
        self._init_alarm_type_to_rule()
        self._cache_all_alarms()

    def _init_aodh_event_actions(self):
        self.actions = {
            AodhEventType.CREATION: self._convert_alarm_creation_event,
            AodhEventType.RULE_CHANGE: self._convert_alarm_rule_change_event,
            AodhEventType.STATE_TRANSITION:
                self._convert_alarm_state_transition_event,
            AodhEventType.DELETION: self._convert_alarm_deletion_event
        }

    def _init_convert_aodh_alarm_rule_actions(self):
        self.convert_rule_actions = {
            AodhExtendedAlarmType.VITRAGE:
                self._convert_vitrage_alarm_rule,
            AodhExtendedAlarmType.EVENT:
                self._convert_event_alarm_rule,
            AodhExtendedAlarmType.THRESHOLD:
                self._convert_threshold_alarm_rule,
            AodhExtendedAlarmType.GNOCCHI_RESOURCES_THRESHOLD:
                self._convert_gnocchi_resources_threshold_alarm_rule,
            AodhExtendedAlarmType.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD:
                self._convert_gnocchi_aggregation_by_metrics_threshold_rule,
            AodhExtendedAlarmType.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD:
                self._convert_gnocchi_aggregation_by_resources_threshold_rule,
            AodhExtendedAlarmType.COMPOSITE:
                self._convert_composite_alarm_rule

        }

    def _init_alarm_type_to_rule(self):
        self.alarm_rule_types = {
            AodhExtendedAlarmType.VITRAGE: AodhProps.EVENT_RULE,
            AodhExtendedAlarmType.EVENT: AodhProps.EVENT_RULE,
            AodhExtendedAlarmType.THRESHOLD: AodhProps.THRESHOLD_RULE,
            AodhExtendedAlarmType.GNOCCHI_RESOURCES_THRESHOLD:
                AodhProps.GNOCCHI_RESOURCES_THRESHOLD_RULE,
            AodhExtendedAlarmType.COMPOSITE: AodhProps.COMPOSITE_RULE,
            AodhExtendedAlarmType.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD:
                AodhProps.GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD_RULE,
            AodhExtendedAlarmType.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD:
                AodhProps.GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD_RULE
        }

    @property
    def client(self):
        if not self._client:
            self._client = os_clients.aodh_client()
        return self._client

    def _vitrage_type(self):
        return AODH_DATASOURCE

    def _alarm_key(self, alarm):
        return alarm[AodhProps.ALARM_ID]

    def _cache_all_alarms(self):
        alarms = self._get_alarms()
        self._filter_and_cache_alarms(alarms,
                                      self._filter_get_valid)

    def _get_alarms(self):
        try:
            aodh_alarms = self.client.alarm.list()
            return [self._convert_alarm(alarm) for alarm in
                    aodh_alarms if alarm is not None]
        except Exception:
            LOG.exception("Failed to get all alarms.")
        return []

    def _is_erroneous(self, alarm):
        return alarm and alarm[AodhProps.STATE] == AodhState.ALARM

    def _status_changed(self, alarm1, alarm2):
        return alarm1 and alarm2 and \
            not alarm1[AodhProps.STATE] == alarm2[AodhProps.STATE]

    def _is_valid(self, alarm):
        return True

    def _get_aodh_alarm_type(self, alarm):

        Aodh_type = [AodhExtendedAlarmType.EVENT,
                     AodhExtendedAlarmType.THRESHOLD,
                     AodhExtendedAlarmType.GNOCCHI_RESOURCES_THRESHOLD,
                     AodhExtendedAlarmType.
                     GNOCCHI_AGGREGATION_BY_METRICS_THRESHOLD,
                     AodhExtendedAlarmType.
                     GNOCCHI_AGGREGATION_BY_RESOURCES_THRESHOLD,
                     AodhExtendedAlarmType.COMPOSITE]

        alarm_type = alarm[AodhProps.TYPE]
        if alarm_type == AodhProps.EVENT and \
                _is_vitrage_alarm(alarm.get(AodhProps.EVENT_RULE)
                                  or alarm.get(AodhProps.RULE)):
            return AodhExtendedAlarmType.VITRAGE
        elif alarm_type not in Aodh_type:
            LOG.warning('Unsupported Aodh alarm of type %s' % alarm_type)
            alarm_type = None

        return alarm_type

    def _convert_alarm(self, alarm):

        entity = self._convert_alarm_common(alarm)

        detail = self._convert_alarm_detail(alarm)
        entity.update(detail)

        alarm_type = self._get_aodh_alarm_type(alarm)
        alarm_rule_type = self.alarm_rule_types[alarm_type]
        alarm_rule = alarm[alarm_rule_type]
        rule = self._convert_alarm_rule(alarm_type, alarm_rule)
        entity.update(rule)

        return entity

    @staticmethod
    def get_event_types():
        # Add event_types to receive notifications about
        return [AodhEventType.CREATION,
                AodhEventType.STATE_TRANSITION,
                AodhEventType.RULE_CHANGE,
                AodhEventType.DELETION]

    def enrich_event(self, event, event_type):
        if event_type in self.actions:
            entity = self.actions[event_type](event)
        else:
            LOG.warning('Unsupported Aodh event type %s' % event_type)
            return None

        # Don't need to update entity, only update the cache
        if entity is None:
            return None

        entity[DSProps.EVENT_TYPE] = event_type

        return AodhDriver.make_pickleable([entity],
                                          AODH_DATASOURCE,
                                          DatasourceAction.UPDATE)[0]

    @classmethod
    def _convert_vitrage_alarm_rule(cls, rule):
        return {
            AodhProps.VITRAGE_ID: _parse_query(rule, AodhProps.VITRAGE_ID),
            AodhProps.RESOURCE_ID: _parse_query(rule, AodhProps.RESOURCE_ID)
        }

    @classmethod
    def _convert_threshold_alarm_rule(cls, rule):
        return cls._alarm_rule_common(rule)

    @classmethod
    def _convert_gnocchi_resources_threshold_alarm_rule(cls, rule):
        return {
            AodhProps.RESOURCE_ID: rule[AodhProps.RESOURCE_ID]
        }

    @classmethod
    def _convert_gnocchi_aggregation_by_metrics_threshold_rule(cls, rule):
        return cls._alarm_rule_common(rule)

    @classmethod
    def _convert_gnocchi_aggregation_by_resources_threshold_rule(cls, rule):
        return cls._alarm_rule_common(rule)

    @classmethod
    def _convert_composite_alarm_rule(cls, rule):
        return cls._alarm_rule_common(rule)

    @classmethod
    def _alarm_rule_common(cls, rule):
        return {
            AodhProps.RESOURCE_ID: _parse_query(rule, AodhProps.RESOURCE_ID)
        }

    @classmethod
    def _convert_event_alarm_rule(cls, rule):
        return {
            AodhProps.EVENT_TYPE: rule[AodhProps.EVENT_TYPE],
            AodhProps.RESOURCE_ID:
                _parse_query(rule, AodhProps.EVENT_RESOURCE_ID)
        }

    @classmethod
    def _convert_alarm_common(cls, alarm):
        return {
            AodhProps.ALARM_ID: alarm.get(AodhProps.ALARM_ID),
            AodhProps.USER_ID: alarm.get(AodhProps.USER_ID),
            AodhProps.PROJECT_ID: alarm.get(AodhProps.PROJECT_ID),
            AodhProps.SEVERITY: alarm.get(AodhProps.SEVERITY),
            AodhProps.TIMESTAMP: alarm.get(AodhProps.TIMESTAMP)
        }

    @classmethod
    def _convert_alarm_detail(cls, alarm):

        return {
            AodhProps.DESCRIPTION: alarm.get(AodhProps.DESCRIPTION),
            AodhProps.ENABLED: alarm.get(AodhProps.ENABLED),
            AodhProps.NAME: alarm.get(AodhProps.NAME),
            AodhProps.STATE: alarm.get(AodhProps.STATE),
            AodhProps.REPEAT_ACTIONS: alarm.get(AodhProps.REPEAT_ACTIONS),
            AodhProps.TYPE: alarm.get(AodhProps.TYPE),
            AodhProps.STATE_TIMESTAMP: alarm.get(AodhProps.STATE_TIMESTAMP),
            AodhProps.STATE_REASON: alarm.get(AodhProps.STATE_REASON)
        }

    def _convert_alarm_rule(self, alarm_type, alarm_rule):
        if alarm_type in self.convert_rule_actions:
            entity = self.convert_rule_actions[alarm_type](alarm_rule)
        else:
            LOG.warning('Unsupported Aodh alarm type %s' % alarm_type)
            return None

        return entity

    @classmethod
    def _parse_changed_rule(cls, change_rule):
        entity = {}
        if AodhProps.EVENT_TYPE in change_rule:
            entity[AodhProps.EVENT_TYPE] = change_rule[AodhProps.EVENT_TYPE]
        if 'query' in change_rule:
            event_resource_id = \
                _parse_query(change_rule, AodhProps.EVENT_RESOURCE_ID)
            resource_id = \
                _parse_query(change_rule, AodhProps.RESOURCE_ID)
            if event_resource_id or resource_id:
                entity[AodhProps.RESOURCE_ID] = event_resource_id if \
                    event_resource_id is not None else resource_id

        return entity

    @staticmethod
    def should_delete_outdated_entities():
        return True

    def _convert_alarm_creation_event(self, event):
        entity = self._convert_alarm_common(event)

        alarm_info = event[AodhProps.DETAIL]
        detail = self._convert_alarm_detail(alarm_info)
        entity.update(detail)

        alarm_type = self._get_aodh_alarm_type(alarm_info)
        alarm_rule = alarm_info[AodhProps.RULE]
        rule_info = self._convert_alarm_rule(alarm_type, alarm_rule)
        entity.update(rule_info)

        return self._filter_and_cache_alarm(entity, None,
                                            self._filter_get_erroneous,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_rule_change_event(self, event):
        """handle alarm rule change notification

        example of changed rule:
        "detail": {"severity": "critical",
                   "rule":
                       {"query": [{"field": "traits.resource_id",
                                   "type": "",
                                   "value": "1",
                                   "op": "eq"}],
                       "event_type": "instance.update"}}
        """
        old_alarm = self._old_alarm(event)
        entity = old_alarm.copy()
        changed_rule = event[AodhProps.DETAIL]
        for (changed_type, changed_info) in changed_rule.items():
            # handle changed rule which may effect the neighbor
            if changed_type == AodhProps.RULE:
                entity.update(self._parse_changed_rule(
                    changed_rule[changed_type]))
            # handle other changed alarm properties
            elif changed_type in vars(AodhProps).values():
                entity[changed_type] = changed_info

        return self._filter_and_cache_alarm(entity, old_alarm,
                                            self._filter_get_erroneous,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_state_transition_event(self, event):

        old_alarm = self._old_alarm(event)
        entity = old_alarm.copy()
        try:
            entity[AodhProps.STATE] = event[AodhProps.DETAIL][AodhProps.STATE]
        except Exception:
            LOG.exception("Failed to Convert alarm state transition event.")

        return self._filter_and_cache_alarm(entity, old_alarm,
                                            self._filter_get_change,
                                            datetime_utils.utcnow(False))

    def _convert_alarm_deletion_event(self, event):
        alarm_key = self._alarm_key(event)
        alarm = self.cache.pop(alarm_key)[0]
        return alarm if self._is_erroneous(alarm) else None


def _parse_query(data, key):
    """Find the relevant key in a given alarm detail query.

    :param data: A query is either a list of this form:
    [
        {
        field: resource_id,
        value: 54132
        }
    ]
    or a string-represented dict of this form:
    '{
        =: { resource_id : 1235423 }
     }'
    """

    query_fields = data.get(AodhProps.QUERY, {})
    try:
        if isinstance(query_fields, str):
            query_fields = json.loads(query_fields)
        if not isinstance(query_fields, list):
            query_fields = [query_fields]
        for query in query_fields:
            field = query.get('field')
            if field and field == key:
                return query['value']
            elif not field:
                field = query.get('=', {})
                for k in field:
                    if k == key:
                        return field[key]
        return None

    except Exception:
        LOG.exception("Failed to parse AODH alarm query")
        return None


def _is_vitrage_alarm(rule):
    return _parse_query(rule, AodhProps.VITRAGE_ID) is not None



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.datasources.nagios.config import NagiosConfig
from vitrage.datasources.nagios.config import NagiosHostMapping
from vitrage.datasources.nagios import NAGIOS_DATASOURCE
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.tests import base
from vitrage.tests.mocks import utils


class TestNagiosConfig(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.TRANSFORMER,
                   default='vitrage.datasources.nagios.transformer.'
                           'NagiosTransformer',
                   help='Nagios data source transformer class path',
                   required=True),
        cfg.StrOpt(DSOpts.DRIVER,
                   default='vitrage.datasources.nagios.driver.NagiosDriver',
                   help='Nagios driver class path',
                   required=True),
        cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                   default=30,
                   min=30,
                   help='interval between checking changes in nagios plugin',
                   required=True),
        cfg.StrOpt('user', default='nagiosadmin',
                   help='Nagios user name'),
        cfg.StrOpt('password', default='nagiosadmin',
                   help='Nagios user password'),
        cfg.StrOpt('url', default='', help='Nagios url'),
        cfg.StrOpt(DSOpts.CONFIG_FILE,
                   default=utils.get_resources_dir() +
                   '/nagios/nagios_conf.yaml',
                   help='Nagios configuration file'),
    ]

    # the mappings match the ones in nagios_conf.yaml
    MAPPING_1 = NagiosHostMapping('compute-1',
                                  NOVA_HOST_DATASOURCE,
                                  'compute-1')
    MAPPING_2 = NagiosHostMapping('compute-2', NOVA_HOST_DATASOURCE, 'host2')
    MAPPING_3 = NagiosHostMapping('compute-(.*)',
                                  NOVA_HOST_DATASOURCE,
                                  '${nagios_host}')
    MAPPING_4 = NagiosHostMapping('instance-(.*)',
                                  NOVA_INSTANCE_DATASOURCE,
                                  '${nagios_host}')
    MAPPINGS = [MAPPING_1, MAPPING_2, MAPPING_3, MAPPING_4]

    NON_EXISTING_MAPPING_1 = NagiosHostMapping('X',
                                               NOVA_HOST_DATASOURCE,
                                               'compute-1')
    NON_EXISTING_MAPPING_2 = NagiosHostMapping('compute-1',
                                               'X',
                                               'compute-1')
    NON_EXISTING_MAPPING_3 = NagiosHostMapping('compute-1',
                                               NOVA_HOST_DATASOURCE,
                                               'X')
    NON_EXISTING_MAPPINGS = [NON_EXISTING_MAPPING_1,
                             NON_EXISTING_MAPPING_2,
                             NON_EXISTING_MAPPING_3]

    def setUp(self):
        super(TestNagiosConfig, self).setUp()
        self.conf_reregister_opts(self.OPTS, NAGIOS_DATASOURCE)

    def test_nagios_configuration_loading(self):
        # Action
        nagios_conf = NagiosConfig()

        # Test assertions
        mappings = nagios_conf.mappings
        self.assertIsNotNone(nagios_conf, "no nagios configuration loaded")
        self.assertEqual(len(self.MAPPINGS), len(mappings))

        for expected_mapping in self.MAPPINGS:
            self.assertTrue(TestNagiosConfig._check_contains(expected_mapping,
                                                             mappings))
        for expected_mapping in self.NON_EXISTING_MAPPINGS:
            self.assertFalse(TestNagiosConfig._check_contains(expected_mapping,
                                                              mappings))

    def test_nagios_mapping(self):
        # check non-regexp mapping
        mapped_resource = self.MAPPING_1.map(None)
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = self.MAPPING_1.map('')
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = self.MAPPING_1.map('compute-1')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(self.MAPPING_1.resource_type, mapped_resource[0])
        self.assertEqual(self.MAPPING_1.resource_name, mapped_resource[1])

        mapped_resource = self.MAPPING_1.map('compute-2')
        self.assertIsNone(mapped_resource, 'expected None')

        # check mapping to a different resource name
        mapped_resource = self.MAPPING_2.map('compute-2')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(self.MAPPING_2.resource_type, mapped_resource[0])
        self.assertEqual(self.MAPPING_2.resource_name, mapped_resource[1])

        # check regexp mapping
        mapped_resource = self.MAPPING_3.map(None)
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = self.MAPPING_3.map('')
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = self.MAPPING_3.map('compute8')
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = self.MAPPING_3.map('compute-8')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(self.MAPPING_3.resource_type, mapped_resource[0])
        self.assertEqual('compute-8', mapped_resource[1])

        mapped_resource = self.MAPPING_3.map('instance-8')
        self.assertIsNone(mapped_resource, 'expected None')

    def test_get_vitrage_resource(self):
        """Test the resource returned after processing a list of mappings

        :return:
        """
        # Action
        nagios_conf = NagiosConfig()

        # Test assertions
        mapped_resource = nagios_conf.get_vitrage_resource(None)
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = nagios_conf.get_vitrage_resource('')
        self.assertIsNone(mapped_resource, 'expected None')

        mapped_resource = nagios_conf.get_vitrage_resource('compute-1')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(NOVA_HOST_DATASOURCE, mapped_resource[0])
        self.assertEqual('compute-1', mapped_resource[1])

        mapped_resource = nagios_conf.get_vitrage_resource('compute-2')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(NOVA_HOST_DATASOURCE, mapped_resource[0])
        self.assertEqual('host2', mapped_resource[1])

        mapped_resource = nagios_conf.get_vitrage_resource('compute-88')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(NOVA_HOST_DATASOURCE, mapped_resource[0])
        self.assertEqual('compute-88', mapped_resource[1])

        mapped_resource = nagios_conf.get_vitrage_resource('instance-7')
        self.assertIsNotNone(mapped_resource, 'expected Not None')
        self.assertEqual(NOVA_INSTANCE_DATASOURCE, mapped_resource[0])
        self.assertEqual('instance-7', mapped_resource[1])

    @staticmethod
    def _check_contains(expected_mapping, mappings):
        for mapping in mappings:
            if TestNagiosConfig._assert_equals(expected_mapping, mapping):
                return True
        return False

    @staticmethod
    def _assert_equals(mapping1, mapping2):
        return mapping1.nagios_host_regexp.pattern == \
            mapping2.nagios_host_regexp.pattern and \
            mapping1.resource_type == mapping2.resource_type and \
            mapping1.resource_name == mapping2.resource_name



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from vitrage.common.constants import NotifierEventTypes
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources import NOVA_HOST_DATASOURCE
from vitrage.datasources import NOVA_INSTANCE_DATASOURCE
from vitrage.notifier.plugins.base import NotifierBase
from vitrage.notifier.plugins.nova import InstanceState
from vitrage import os_clients

LOG = logging.getLogger(__name__)


class NovaNotifier(NotifierBase):

    @staticmethod
    def get_notifier_name():
        return 'nova'

    def __init__(self):
        super(NovaNotifier, self).__init__()
        self.client = os_clients.nova_client()
        self.actions = {
            NOVA_HOST_DATASOURCE: self._mark_host_down,
            NOVA_INSTANCE_DATASOURCE: self._reset_instance_state
        }

    def process_event(self, data, event_type):
        if not data or not event_type:
            return

        if event_type == NotifierEventTypes.ACTIVATE_MARK_DOWN_EVENT or \
                event_type == NotifierEventTypes.DEACTIVATE_MARK_DOWN_EVENT:

            is_down = event_type == NotifierEventTypes.ACTIVATE_MARK_DOWN_EVENT
            action = self.actions.get(data.get(VProps.VITRAGE_TYPE))

            if action:
                action(data.get(VProps.ID), is_down)
            else:
                LOG.warning('Unsupported datasource type %s for mark_down '
                            'action', data.get(VProps.VITRAGE_TYPE))

    def _mark_host_down(self, host_id, is_down):
        try:
            LOG.info('Nova services.force_down - host id: %s, is_down: %s',
                     host_id, is_down)
            response = self.client.services.force_down(
                host_id, 'nova-compute', is_down)
            LOG.info('RESPONSE %s', response.to_dict())
        except Exception:
            LOG.exception('Failed to services.force_down.')

    def _reset_instance_state(self, server_id, is_down):
        state = InstanceState.ERROR if is_down else InstanceState.ACTIVE
        try:
            LOG.info('Nova servers.reset_state - server: %s, state: %s',
                     server_id, state)
            response = self.client.servers.reset_state(server_id, state)
            LOG.info('RESPONSE %s', response)
        except Exception:
            LOG.exception('Failed to execute servers.reset_state.')



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class HandlerBase(object, metaclass=abc.ABCMeta):

    def __init__(self):
        pass

    @abc.abstractmethod
    def undefined_property(self):
        pass

    @abc.abstractmethod
    def value_properties(self):
        pass

    @abc.abstractmethod
    def set_aggregated_value(self, new_vertex, aggregated_value):
        pass

    @abc.abstractmethod
    def set_operational_value(self, new_vertex, operational_value):
        pass

    @abc.abstractmethod
    def default_values(self):
        pass

    @abc.abstractmethod
    def get_value_class_instance(self):
        pass



# Copyright 2020 - Inspur - Qitao
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps

from vitrage.datasources.resource_transformer_base import \
    ResourceTransformerBase
from vitrage.datasources.transformer_base import extract_field_value
import vitrage.graph.utils as graph_utils

from vitrage.datasources import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources import transformer_base as tbase

from vitrage.datasources.cetus.pod import CETUS_POD_DATASOURCE
from vitrage.datasources.cetus.properties import CetusPodProperties \
    as CetusProp

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PodTransformer(ResourceTransformerBase):

    def _create_vertex(self, entity_event, state, node_name):
        metadata = {
            VProps.NAME: node_name
        }

        entity_key = self._create_entity_key(entity_event)
        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]
        update_timestamp = self._format_update_timestamp(
            extract_field_value(entity_event, DSProps.SAMPLE_DATE),
            vitrage_sample_timestamp)

        return graph_utils.create_vertex(
            entity_key,
            vitrage_category=EntityCategory.RESOURCE,
            vitrage_type=CETUS_POD_DATASOURCE,
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            update_timestamp=update_timestamp,
            entity_id=extract_field_value(entity_event, CetusProp.ID),
            entity_state=state,
            metadata=metadata
        )

    def _create_snapshot_entity_vertex(self, entity_event):
        node_name = extract_field_value(entity_event, CetusProp.NAME)
        state = extract_field_value(entity_event, CetusProp.STATUS)
        return self._create_vertex(entity_event, state, node_name)

    def _create_update_entity_vertex(self, entity_event):
        node_name = extract_field_value(entity_event, CetusProp.NAME)
        state = extract_field_value(entity_event, CetusProp.STATUS)
        return self._create_vertex(entity_event, state, node_name)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_pod_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_pod_neighbors(entity_event)

    def _create_entity_key(self, event):
        instance_id = extract_field_value(event, CetusProp.ID)
        key_fields = self._key_values(
            CETUS_POD_DATASOURCE, instance_id)
        key = tbase.build_key(key_fields)
        return key

    def get_vitrage_type(self):
        return CETUS_POD_DATASOURCE

    def _create_pod_neighbors(self, entity_event):
        neighbors = []
        node = extract_field_value(entity_event, CetusProp.NODE)

        node_neighbor = self._create_neighbor(
            entity_event,
            node,
            NOVA_INSTANCE_DATASOURCE,
            EdgeLabel.CONTAINS,
            is_entity_source=False,
        )
        neighbors.append(node_neighbor)
        return neighbors



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_config import cfg
from oslo_log import log

from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.entity_graph.mappings.alarm_handler import AlarmHandler
from vitrage.entity_graph.mappings.resource_handler import \
    ResourceHandler
from vitrage.utils import file as file_utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


DEFAULT_INFO_MAPPER = 'default'


class DatasourceInfoMapper(object):
    OPERATIONAL_VALUES = 'operational_values'
    PRIORITY_VALUES = 'priority_values'
    UNDEFINED_DATASOURCE = 'undefined datasource'

    def __init__(self):
        self.category_normalizer = self._init_category_normalizer()
        self.datasources_value_confs = self._load_value_configurations()

    def vitrage_operational_value(self, vitrage_type, value):
        return self._get_value_data(vitrage_type,
                                    value,
                                    self.OPERATIONAL_VALUES)

    def value_priority(self, vitrage_type, value):
        return self._get_value_data(vitrage_type,
                                    value,
                                    self.PRIORITY_VALUES)

    def vitrage_aggregate_values(self, new_vertex, graph_vertex):
        LOG.debug('new_vertex: %s', new_vertex)
        LOG.debug('graph_vertex: %s', graph_vertex)

        vitrage_type = new_vertex[VProps.VITRAGE_TYPE] if \
            VProps.VITRAGE_TYPE in new_vertex.properties else \
            graph_vertex[VProps.VITRAGE_TYPE]

        vitrage_category = new_vertex[VProps.VITRAGE_CATEGORY] if \
            VProps.VITRAGE_CATEGORY in new_vertex.properties else \
            graph_vertex[VProps.VITRAGE_CATEGORY]

        if vitrage_type in self.datasources_value_confs or \
                vitrage_type not in CONF.datasources.types:
            value_properties = \
                self.category_normalizer[vitrage_category].value_properties()
            vitrage_oper_value, vitrage_aggregated_value, value_priority = \
                self._find_operational_value_and_priority(new_vertex,
                                                          graph_vertex,
                                                          value_properties[0],
                                                          vitrage_type)
            value_properties.pop(0)

            for property_ in value_properties:
                t_operational_value, t_aggregated_value, t_value_priority = \
                    self._find_operational_value_and_priority(new_vertex,
                                                              graph_vertex,
                                                              property_,
                                                              vitrage_type)
                if t_value_priority > value_priority:
                    vitrage_oper_value = t_operational_value
                    vitrage_aggregated_value = t_aggregated_value
                    value_priority = t_value_priority

            self.category_normalizer[vitrage_category].set_aggregated_value(
                new_vertex, vitrage_aggregated_value)
            self.category_normalizer[vitrage_category].set_operational_value(
                new_vertex, vitrage_oper_value)
        else:
            self.category_normalizer[vitrage_category].set_aggregated_value(
                new_vertex, self.UNDEFINED_DATASOURCE)
            self.category_normalizer[vitrage_category].set_operational_value(
                new_vertex, self.UNDEFINED_DATASOURCE)

    def get_datasource_priorities(self, vitrage_type=None):
        if vitrage_type:
            datasource_info = self.datasources_value_confs[vitrage_type]
            return datasource_info[self.PRIORITY_VALUES]
        else:
            priorities_dict = \
                {key: self.datasources_value_confs[key][self.PRIORITY_VALUES]
                 for key in self.datasources_value_confs.keys()}
            return priorities_dict

    @staticmethod
    def _init_category_normalizer():
        return {
            EntityCategory.RESOURCE: ResourceHandler(),
            EntityCategory.ALARM: AlarmHandler()
        }

    def _load_value_configurations(self):
        valid_datasources_conf = {}
        erroneous_datasources_conf = []

        files = file_utils.list_files(
            CONF.entity_graph.datasources_values_dir, '.yaml')

        for file_name in files:
            try:
                full_path = CONF.entity_graph.datasources_values_dir \
                    + '/' + file_name
                operational_values, priority_values = \
                    self._retrieve_values_and_priorities_from_file(full_path)
                valid_datasources_conf[os.path.splitext(file_name)[0]] = {
                    self.OPERATIONAL_VALUES: operational_values,
                    self.PRIORITY_VALUES: priority_values
                }
            except Exception:
                datasource = os.path.splitext(file_name)[0]
                erroneous_datasources_conf.append(datasource)
                LOG.exception('Erroneous data source is %s', datasource)

        self._check_value_confs_exists(
            [key for key in valid_datasources_conf.keys()],
            erroneous_datasources_conf)

        return valid_datasources_conf

    def _retrieve_values_and_priorities_from_file(self, full_path):
        values = {}
        priorities = {}
        config = file_utils.load_yaml_file(full_path, with_exception=True)
        vitrage_category = config['category']

        for item in config['values']:
            aggregated_values = item['aggregated values']
            priority_value = int(aggregated_values['priority'])

            # original to operational value
            for value_map in aggregated_values['original values']:
                name = value_map['name']
                operational_value = value_map['operational_value']
                values[name.upper()] = operational_value
                priorities[name.upper()] = priority_value

        self._check_validity(vitrage_category, values, priorities, full_path)

        self._add_default_values(values, priorities, vitrage_category)

        return values, priorities

    def _add_default_values(self, values, priorities, category):
        default_values = self.category_normalizer[category].default_values()
        for original_val, operational_val, priority_val in default_values:
            values[original_val] = operational_val
            priorities[original_val] = priority_val

    def _check_validity(self, category, values, priorities, full_path):
        # check that all the operational values exists
        state_class_instance = \
            self.category_normalizer[category].get_value_class_instance()
        operational_values = DatasourceInfoMapper.\
            _get_all_local_variables_of_class(state_class_instance)
        for operational_value in values.values():
            if operational_value not in operational_values:
                raise ValueError('operational value %s for %s is not in %s',
                                 operational_value, full_path,
                                 state_class_instance.__class__.__name__)

    def _get_value_data(self, vitrage_type, value, data_type):
        try:
            upper_value = value if not value else value.upper()

            if vitrage_type in self.datasources_value_confs:
                values_conf = self.datasources_value_confs[
                    vitrage_type][data_type]

                return values_conf[upper_value] if upper_value in values_conf \
                    else values_conf[None]
            else:
                values_conf = self.datasources_value_confs[
                    DEFAULT_INFO_MAPPER][data_type]

                return values_conf[upper_value] if upper_value in values_conf \
                    else values_conf[None]
        except Exception:
            LOG.error('Exception in datasource: %s', vitrage_type)
            raise

    def _find_operational_value_and_priority(self,
                                             new_vertex,
                                             graph_vertex,
                                             property_,
                                             vitrage_type):
        state = self._get_updated_property(new_vertex,
                                           graph_vertex,
                                           property_)

        upper_state = state if not state else state.upper()

        vitrage_operational_state = self.vitrage_operational_value(
            vitrage_type, upper_state)

        value_priority = self.value_priority(vitrage_type,
                                             upper_state)

        return vitrage_operational_state, upper_state, value_priority

    @staticmethod
    def _get_all_local_variables_of_class(class_instance):
        return [getattr(class_instance, attr) for attr in dir(class_instance)
                if not callable(attr) and not attr.startswith("__")]

    def _check_value_confs_exists(self,
                                  ok_datasources,
                                  error_datasources):

        datasource_types = CONF.datasources.types
        datasources_with_state_conf = ok_datasources + error_datasources

        for datasource_type in datasource_types:
            if datasource_type not in datasources_with_state_conf:
                LOG.info("No state configuration file for: %s",
                         datasource_type)

    @staticmethod
    def _get_updated_property(new_vertex, graph_vertex, prop):
        if new_vertex and prop in new_vertex.properties:
            return new_vertex[prop]
        elif graph_vertex and prop in graph_vertex.properties:
            return graph_vertex[prop]

        return None



# Copyright 2019 - Viettel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class KapacitorProperties(object):
    ID = 'id'
    RESOURCE_TYPE = 'resource_type'
    RESOURCE_NAME = 'resource_name'
    DETAILS = 'details'
    STATUS = 'status'
    HOST = 'host'
    PRIORITY = 'priority'
    TIME = 'time'
    MESSAGE = 'message'


class KapacitorState(object):
    OK = 'ok'
    INFO = 'info'
    WARNING = 'warning'
    CRITICAL = 'critical'



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import VertexProperties as VProps
from vitrage.entity_graph.mappings.datasource_info_mapper \
    import DEFAULT_INFO_MAPPER
from vitrage.evaluator.template_fields import TemplateFields


class BaselineTools(object):
    @staticmethod
    def get_score(action_info):
        return 1  # no priorities

    @classmethod
    def get_extra_info(cls, action_specs):
        return None


class RaiseAlarmTools(object):

    def __init__(self, scores):
        self.scores = scores

    def get_score(self, action_info):
        severity = action_info.specs.properties[TemplateFields.SEVERITY]
        return self.scores.get(severity.upper(), 0)

    @classmethod
    def get_extra_info(cls, action_specs):
        return action_specs.properties[TemplateFields.ALARM_NAME]


class SetStateTools(object):

    def __init__(self, scores):
        self.scores = scores

    def get_score(self, action_info):
        state = action_info.specs.properties[TemplateFields.STATE].upper()
        target_resource = action_info.specs.targets[TemplateFields.TARGET]
        target_vitrage_type = target_resource[VProps.VITRAGE_TYPE]
        score_name = target_vitrage_type \
            if target_vitrage_type in self.scores else DEFAULT_INFO_MAPPER
        return self.scores[score_name].get(state, 0)

    @classmethod
    def get_extra_info(cls, action_specs):
        return None



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from operator import itemgetter

from oslo_log import log
import pecan

from pecan.core import abort
from vitrage.api.controllers.rest import RootRestController
from vitrage.api.policy import enforce

LOG = log.getLogger(__name__)


# noinspection PyBroadException
class ServiceController(RootRestController):
    @pecan.expose('json')
    def index(self):
        return self.get()

    @pecan.expose('json')
    def get(self):
        enforce("get service list", pecan.request.headers,
                pecan.request.enforcer, {})

        LOG.info('received get service list')

        coordinator = pecan.request.coordinator
        if not coordinator.backend_url:
            abort(500, 'Service API not supported')
        if not coordinator.is_active():
            abort(500, 'Failed to connect to coordination backend')

        try:
            return sorted(coordinator.get_services(), key=itemgetter('name'))
        except Exception:
            LOG.exception('failed to get service list.')
            abort(404, 'Failed to get service list.')



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

ADD_VERTEX = 'add_vertex'
REMOVE_VERTEX = 'remove_vertex'
UPDATE_VERTEX = 'update_vertex'
ADD_EDGE = 'add_edge'
REMOVE_EDGE = 'remove_edge'
EXECUTE_EXTERNAL = 'execute_external'

EXECUTION_ENGINE = 'execution_engine'



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging

from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EdgeProperties as EProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.exception import VitrageTransformerError
from vitrage.datasources import transformer_base
from vitrage.datasources.transformer_base import Neighbor
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.evaluator.actions.recipes.action_steps import ADD_EDGE
from vitrage.evaluator.actions.recipes.action_steps import ADD_VERTEX
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_EDGE
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_VERTEX
from vitrage.evaluator.actions.recipes.action_steps import UPDATE_VERTEX
from vitrage.evaluator.actions.recipes.base import EVALUATOR_EVENT_TYPE
from vitrage.evaluator.template_fields import TemplateFields as TFields
import vitrage.graph.utils as graph_utils
from vitrage.graph import Vertex


LOG = logging.getLogger(__name__)


VITRAGE_DATASOURCE = 'vitrage'


class EvaluatorEventTransformer(transformer_base.TransformerBase):

    def __init__(self, transformers):
        super(EvaluatorEventTransformer, self).__init__(transformers)
        self.actions = self._init_actions()

    def _create_snapshot_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_vertex(self, event):

        event_type = event[EVALUATOR_EVENT_TYPE]

        update_timestamp = transformer_base.convert_timestamp_format(
            '%Y-%m-%d %H:%M:%S.%f',
            event[VProps.UPDATE_TIMESTAMP]
        )

        if event_type == UPDATE_VERTEX:
            properties = {
                VProps.UPDATE_TIMESTAMP: update_timestamp,
                VProps.VITRAGE_IS_PLACEHOLDER: False,
                VProps.RESOURCE_ID: event.get(TFields.TARGET)
            }
            if VProps.IS_REAL_VITRAGE_ID in event:
                properties[VProps.IS_REAL_VITRAGE_ID] = \
                    event.get(VProps.IS_REAL_VITRAGE_ID)
            if VProps.VITRAGE_STATE in event:
                properties[VProps.VITRAGE_STATE] = \
                    event.get(VProps.VITRAGE_STATE)
            if VProps.IS_MARKED_DOWN in event:
                properties[VProps.IS_MARKED_DOWN] = \
                    event.get(VProps.IS_MARKED_DOWN)

            return Vertex(event[VProps.VITRAGE_ID], properties)

        if event_type in [ADD_VERTEX, REMOVE_VERTEX]:

            metadata = {
                VProps.NAME: event[TFields.ALARM_NAME],
                VProps.SEVERITY: event[TFields.SEVERITY],
                VProps.STATE: event[VProps.STATE],
                VProps.RESOURCE_ID: event.get(TFields.TARGET)
            }
            return graph_utils.create_vertex(
                self._create_entity_key(event),
                vitrage_category=EntityCategory.ALARM,
                vitrage_type=VITRAGE_DATASOURCE,
                vitrage_sample_timestamp=event[
                    VProps.VITRAGE_SAMPLE_TIMESTAMP],
                update_timestamp=update_timestamp,
                metadata=metadata)

        return None

    def _create_snapshot_neighbors(self, event):
        return self._create_vitrage_neighbors(event)

    def _create_update_neighbors(self, event):
        return self._create_vitrage_neighbors(event)

    def _create_vitrage_neighbors(self, event):
        event_type = event[EVALUATOR_EVENT_TYPE]

        timestamp = transformer_base.convert_timestamp_format(
            '%Y-%m-%d %H:%M:%S.%f',
            event[VProps.UPDATE_TIMESTAMP]
        )

        if event_type in [ADD_EDGE, REMOVE_EDGE]:

            relation_edge = graph_utils.create_edge(
                source_id=event[TFields.SOURCE],
                target_id=event[TFields.TARGET],
                relationship_type=event[EProps.RELATIONSHIP_TYPE],
                update_timestamp=timestamp)

            return [Neighbor(None, relation_edge)]

        if event_type == ADD_VERTEX:
            result = []

            relation_edge = graph_utils.create_edge(
                source_id=TransformerBase.uuid_from_deprecated_vitrage_id(
                    self._create_entity_key(event)),
                target_id=event[TFields.TARGET],
                relationship_type=EdgeLabel.ON,
                update_timestamp=timestamp)

            neighbor_props = {
                VProps.VITRAGE_IS_PLACEHOLDER: True,
                VProps.UPDATE_TIMESTAMP: timestamp,
                VProps.VITRAGE_SAMPLE_TIMESTAMP:
                    event[VProps.VITRAGE_SAMPLE_TIMESTAMP],
                VProps.IS_REAL_VITRAGE_ID: True,
                VProps.VITRAGE_TYPE: event.get(VProps.VITRAGE_RESOURCE_TYPE),
                VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            }
            result.append(Neighbor(
                Vertex(event[TFields.TARGET], neighbor_props), relation_edge))
            if event.get(TFields.CAUSING_ALARM):
                relation_edge = graph_utils.create_edge(
                    source_id=event[TFields.CAUSING_ALARM],
                    target_id=TransformerBase.uuid_from_deprecated_vitrage_id(
                        self._create_entity_key(event)),
                    relationship_type=EdgeLabel.CAUSES,
                    update_timestamp=timestamp)
                result.append(
                    Neighbor(Vertex(
                        event[TFields.CAUSING_ALARM],
                        {
                            VProps.UPDATE_TIMESTAMP: timestamp,
                            VProps.VITRAGE_IS_PLACEHOLDER: True,
                        }),
                        relation_edge)
                )
            return result

        return []

    def _extract_graph_action(self, event):
        event_type = event[EVALUATOR_EVENT_TYPE]

        try:
            return self.actions[event_type]
        except Exception:
            raise VitrageTransformerError(
                'Invalid Evaluator event type: (%s)' % event_type)

    @staticmethod
    def _init_actions():
        return {
            UPDATE_VERTEX: GraphAction.UPDATE_ENTITY,
            ADD_VERTEX: GraphAction.CREATE_ENTITY,
            REMOVE_VERTEX: GraphAction.DELETE_ENTITY,
            ADD_EDGE: GraphAction.UPDATE_RELATIONSHIP,
            REMOVE_EDGE: GraphAction.DELETE_RELATIONSHIP
        }

    def _create_entity_key(self, event):
        key_fields = self._key_values(event[TFields.ALARM_NAME],
                                      event[TFields.TARGET])
        return transformer_base.build_key(key_fields)

    def _key_values(self, *args):
        return (EntityCategory.ALARM, VITRAGE_DATASOURCE) + args

    def create_neighbor_placeholder_vertex(self, **kwargs):
        LOG.info('Evaluator does not create placeholders')

    def get_vitrage_type(self):
        return VITRAGE_DATASOURCE



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EdgeProperties
from vitrage.evaluator.actions.recipes.action_steps import ADD_EDGE
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_EDGE
from vitrage.evaluator.actions.recipes import base
from vitrage.evaluator.actions.recipes.base import ActionStepWrapper
from vitrage.evaluator.template_fields import TemplateFields as TFields


class AddCausalRelationship(base.Recipe):
    """Connect two alarms in the graph to indicate one cause other (RCA)

    The 'get_do_recipe' and 'get_undo_recipe' receive action_spec as input.
    The action_spec contains the following fields: type, targets and
    properties. example input:

    action_spec = ActionSpecs('type'= 'add_causal_relationship',
                              'targets'= {target: id, source: id},
                              'properties' = {}
    """

    @staticmethod
    def get_do_recipe(action_spec):

        edge_params = AddCausalRelationship._get_edge_params(
            action_spec.targets)
        add_edge_step = ActionStepWrapper(ADD_EDGE, edge_params)

        return [add_edge_step]

    @staticmethod
    def get_undo_recipe(action_spec):

        edge_params = AddCausalRelationship._get_edge_params(
            action_spec.targets)
        remove_edge_step = ActionStepWrapper(REMOVE_EDGE, edge_params)

        return [remove_edge_step]

    @staticmethod
    def _get_edge_params(params):

        return {
            TFields.SOURCE: params[TFields.SOURCE].vertex_id,
            TFields.TARGET: params[TFields.TARGET].vertex_id,
            EdgeProperties.RELATIONSHIP_TYPE: EdgeLabel.CAUSES
        }



# Copyright 2019 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.evaluator.template_fields import TemplateFields as TField
from vitrage.evaluator.template_loading.v3.action_loader import ActionLoader


class RaiseAlarmLoader(ActionLoader):

    def load(self, action_id, default_target, action_dict, action_type):
        """V3 template action raise_alarm to ActionSpecs transformation

        causing_alarm is replaced with a function, to get the vertex id.
        Example:
        causing_alarm: zabbix_alarm
        Replaced with:
        causing_alarm: get_attr(zabbix_alarm, vitrage_id)

        :param action_id: Unique action identifier
        :param default_target: Is taken from the condition,
         it is used when the action doesn't define a target
        :param action_dict: Action section taken from the template.
        :param action_type: example: set_state/raise_alarm/etc..
        :rtype: ActionSpecs
        """
        if TField.CAUSING_ALARM in action_dict:
            action_dict[TField.CAUSING_ALARM] = \
                'get_attr(%s, vitrage_id)' % action_dict[TField.CAUSING_ALARM]
        return super(RaiseAlarmLoader, self).load(
            action_id, default_target, action_dict, action_type)



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.nova_driver_base import NovaDriverBase


class HostDriver(NovaDriverBase):

    def get_all(self, datasource_action):
        hosts = self.client.services.list(binary='nova-compute')
        return self.make_pickleable(
            [h.to_dict() for h in hosts],
            NOVA_HOST_DATASOURCE,
            datasource_action,
            *self.properties_to_filter_out())

    @staticmethod
    def properties_to_filter_out():
        return ['manager']

    @staticmethod
    def should_delete_outdated_entities():
        return True



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from oslo_log import log
from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_schema_factory import TemplateSchemaFactory
from vitrage.evaluator.template_validation.base import get_correct_result
from vitrage.evaluator.template_validation.base import get_fault_result
from vitrage.evaluator.template_validation.base import get_warning_result
from vitrage.evaluator.template_validation.status_messages import status_msgs


LOG = log.getLogger(__name__)
RESULT_DESCRIPTION = 'Template content validation'


def get_content_correct_result():
    return get_correct_result(RESULT_DESCRIPTION)


def get_content_fault_result(code, msg=None):
    return get_fault_result(RESULT_DESCRIPTION, code, msg)


def get_content_warning_result(code):
    return get_warning_result(RESULT_DESCRIPTION, code)


def validate_template_id(definitions_index, id_to_check):
    if id_to_check not in definitions_index:
        msg = status_msgs[3] + ' template id: %s' % id_to_check
        LOG.error('%s status code: %s' % (msg, 3))
        return get_fault_result(RESULT_DESCRIPTION, 3, msg)

    return get_correct_result(RESULT_DESCRIPTION)


def get_template_schema(template):
    metadata = template.get(TemplateFields.METADATA)

    if metadata is None:
        LOG.error('%s status code: %s' % (status_msgs[62], 62))
        return get_content_fault_result(62), None

    version = metadata.get(TemplateFields.VERSION)
    template_schema = TemplateSchemaFactory().template_schema(version)

    if template_schema:
        return get_content_correct_result(), template_schema
    else:
        LOG.error('%s status code: %s' % (status_msgs[63], 63))
        return get_content_fault_result(63), None


class ActionValidator(object, metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def validate(action, definitions_index):
        """Validate the content of the action.

        :param action: The action to be validated
        :type action: dict

        :param definitions_index: Entities and relationships in the template
        :type definitions_index: dict

        """
        pass



# Copyright 2016 - Nokia Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import timedelta
import json
import time

import pecan
from pytz import utc

from oslo_log import log
from oslo_utils.uuidutils import is_uuid_like
from osprofiler import profiler
from pecan.core import abort

from vitrage.api.controllers.rest import RootRestController
from vitrage.api.policy import enforce
from vitrage.common.constants import TemplateStatus as TStatus
from vitrage.common.exception import VitrageError

LOG = log.getLogger(__name__)

ONE_HOUR = int(timedelta(hours=1).total_seconds())


class TemplateVersionsController(RootRestController):
    @pecan.expose('json')
    def index(self):
        return self.get()

    @pecan.expose('json')
    def get(self):
        LOG.info('get template versions')

        try:
            return pecan.request.client.call(pecan.request.context,
                                             'template_versions')
        except Exception:
            LOG.exception('Failed to get template versions')
            abort(404, 'Failed to get template versions')


@profiler.trace_cls("template controller",
                    info={}, hide_args=False, trace_private=False)
class TemplateController(RootRestController):

    versions = TemplateVersionsController()

    @pecan.expose('json')
    def get_all(self):

        LOG.info('returns template list')

        enforce("template list",
                pecan.request.headers,
                pecan.request.enforcer,
                {})
        try:
            return self._get_templates()
        except Exception:
            LOG.exception('failed to get template list.')
            abort(404, 'Failed to get template list')

    @pecan.expose('json')
    def get(self, _id):

        LOG.info('get template content')

        enforce("template show",
                pecan.request.headers,
                pecan.request.enforcer,
                {})

        try:
            return self._show_template(_id)
        except Exception:
            LOG.exception('Failed to show template %s.', _id)
            abort(404, 'Failed to show template.')

    @pecan.expose('json')
    def delete(self, **kwargs):
        # for backward computability
        values = kwargs['uuid'] if 'uuid'in kwargs else kwargs['id']
        LOG.info("delete template. values: %s", values)
        uuids = self._to_uuids(values)
        LOG.info("delete template. uuids: %s", uuids)

        enforce("template delete",
                pecan.request.headers,
                pecan.request.enforcer,
                {})
        try:
            return self._delete(uuids)
        except Exception:
            LOG.exception('Failed to delete template.')
            abort(404, 'Failed to delete template.')

    @pecan.expose('json')
    def put(self, **kwargs):
        templates = kwargs['templates']
        LOG.info("add template: %s", templates)

        enforce("template add",
                pecan.request.headers,
                pecan.request.enforcer,
                {})
        template_type = kwargs['template_type']
        params = kwargs.get('params')
        overwrite = kwargs.get('overwrite')

        if overwrite:
            names = [
                template[1]['metadata']['name']
                for template in templates
            ]

            uuids = self._to_uuids(names)
            self._delete_templates_and_wait(uuids)

        try:
            return self._add(templates, template_type, params)
        except Exception:
            LOG.exception('Failed to add template.')
            abort(404, 'Failed to add template.')

    @pecan.expose('json')
    def post(self, **kwargs):

        LOG.info('validate template. args: %s', kwargs)

        enforce("template validate",
                pecan.request.headers,
                pecan.request.enforcer,
                {})

        templates = kwargs['templates']
        template_type = kwargs.get('template_type')
        params = kwargs.get('params')

        try:
            return self._validate(templates, template_type, params)
        except Exception:
            LOG.exception('Failed to validate template(s).')
            abort(404, 'Failed to validate template(s).')

    @classmethod
    def _get_templates(cls):
        try:
            templates = pecan.request.storage.templates.query()
            for template in templates:
                template.created_at = utc.localize(template.created_at)
                if template.updated_at:
                    template.updated_at = utc.localize(template.updated_at)
            templates = [t for t in templates if t.status != TStatus.DELETED]
            templates.sort(key=lambda templ: templ.created_at)
            return [cls._db_template_to_dict(t) for t in templates]
        except Exception:
            LOG.exception('Failed to get template list.')
            abort(404, 'Failed to get template list.')

    @staticmethod
    def _show_template(_id):
        uuid = TemplateController._to_uuid(_id)
        try:
            templates = pecan.request.storage.templates.query(uuid=uuid)
            if not templates or templates[0].status == TStatus.DELETED:
                raise VitrageError("Template %s not found", uuid)
            return templates[0].file_content
        except Exception:
            LOG.exception('Failed to show template with uuid: %s ', uuid)
            abort(404, 'Failed to show template.')

    @staticmethod
    def _validate(templates, template_type, params):

        result_json = pecan.request.client.call(pecan.request.context,
                                                'validate_template',
                                                templates=templates,
                                                template_type=template_type,
                                                params=params)
        try:
            return json.loads(result_json)
        except Exception:
            LOG.exception('Failed to open template file(s).')
            abort(404, 'Failed to validate template file.')

    @classmethod
    def _add(cls, templates, template_type, params):
        try:
            results = pecan.request.client.call(
                pecan.request.context,
                'add_template',
                templates=templates,
                template_type=template_type,
                params=params,
            )
            return results
        except Exception:
            LOG.exception('Failed to add template file.')
            abort(404, 'Failed to add template file.')

    @staticmethod
    def _db_template_to_dict(template):
        return {
            "uuid": template.uuid,
            "name": template.name,
            "status": template.status,
            "date": template.created_at,
            "status details": template.status_details,
            "type": template.template_type,
        }

    @staticmethod
    def _delete(uuids):
        try:
            results = pecan.request.client.call(
                pecan.request.context,
                'delete_template',
                uuids=uuids)
            return results
        except Exception:
            LOG.exception('Failed to delete template.')
            abort(404, 'Failed to delete template.')

    def _to_uuids(self, values):
        # if it is a single string
        # make sure I don't iterate
        # on the characters
        if type(values) is not list:
            values = [values]

        return [self._to_uuid(val) for val in values]

    @staticmethod
    def _to_uuid(val):
        if is_uuid(val):
            return val
        template = pecan.request.storage.templates.query_with_status_not(
            status=TStatus.DELETED, name=val)
        if template:
            return template.uuid
        return val

    def _delete_templates_and_wait(self, uuids):
        self._delete(uuids)

        def check_deleted():
            for _id in uuids:
                try:
                    self._show_template(_id)
                except Exception:  # if deleted we get exception
                    pass
                else:
                    return False
            return True

        return wait_for_action_to_end(ONE_HOUR, check_deleted)


def is_uuid(val):
    # unwrap the name or id
    # from the list if in a list
    # [value] --> value
    if type(val) is list:
        val, = val
    return is_uuid_like(val)


def wait_for_action_to_end(timeout, func, **kwargs):
    count = 0
    while count < timeout:
        if func(**kwargs):
            return True
        count += 1
        time.sleep(1)
    return False



# Copyright 2019 - Viettel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.datasources.kapacitor.driver import KapacitorDriver
from vitrage.datasources.kapacitor import KAPACITOR_DATASOURCE
from vitrage.datasources.kapacitor.properties \
    import KapacitorProperties as KProps
from vitrage.datasources.kapacitor.properties \
    import KapacitorState as KState
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver
from vitrage.tests.mocks import utils

# notification alarm input
HOST = 'compute-1'
ALARM_PRIntORIY = 'critical'
ALARM_EVENT_TYPE = 'kapacitor.alarm.critical'

# match result
EXPECTED_RESOURCE_TYPE = 'nova.host'
EXPECTED_RESOURCE_NAME = 'compute1'
EXPECTED_EVENT_PRIORIY = 'critical'
EXPECTED_EVENT_TYPE = 'kapacitor.alarm.critical'


class TestKapacitorDriver(base.BaseTest):
    OPTS = [
        cfg.StrOpt(DSOpts.CONFIG_FILE,
                   help='Kapacitor configuration file',
                   default=utils.get_resources_dir()
                   + '/kapacitor/kapacitor_conf.yaml'),
    ]

    def setUp(self):
        super(TestKapacitorDriver, self).setUp()
        self.conf_reregister_opts(self.OPTS, group=KAPACITOR_DATASOURCE)
        self.driver = KapacitorDriver()

    def test_enrich_event(self):
        # Test event on host
        # Setup
        input_data = {KProps.HOST: 'compute-1',
                      KProps.PRIORITY: 'CPU utilization',
                      ALARM_EVENT_TYPE: KState.CRITICAL}
        expected_data = {DSProps.EVENT_TYPE: KState.CRITICAL,
                         KProps.RESOURCE_NAME: 'compute-1',
                         KProps.RESOURCE_TYPE: 'nova.host',
                         KProps.PRIORITY: 'CPU utilization'}
        event = self._generate_event(input_data[KProps.HOST],
                                     input_data[KProps.PRIORITY])
        # Action
        event = self.driver.enrich_event(event,
                                         input_data[ALARM_EVENT_TYPE])
        # Test assertions
        self._assert_event_equal(event, expected_data)

        # Test event on instance
        # Setup
        input_data = {KProps.HOST: 'node1-vm',
                      KProps.PRIORITY: 'CPU utilization',
                      ALARM_EVENT_TYPE: KState.CRITICAL}
        expected_data = {DSProps.EVENT_TYPE: KState.CRITICAL,
                         KProps.RESOURCE_NAME: 'node1-vm',
                         KProps.RESOURCE_TYPE: 'nova.instance',
                         KProps.PRIORITY: 'CPU utilization'}
        event = self._generate_event(input_data[KProps.HOST],
                                     input_data[KProps.PRIORITY])
        # Action
        event = self.driver.enrich_event(event,
                                         input_data[ALARM_EVENT_TYPE])
        # Test assertions
        self._assert_event_equal(event, expected_data)

    @staticmethod
    def _generate_event(hostname, priority):
        update_vals = {}
        if hostname:
            update_vals[KProps.HOST] = hostname
        if priority:
            update_vals[KProps.PRIORITY] = priority

        generators = mock_driver.simple_kapacitor_alarm_generators(
            update_vals=update_vals)

        return mock_driver.generate_sequential_events_list(generators)[0]

    def _assert_event_equal(self,
                            event1,
                            event2):
        self.assertIsNotNone(event1, 'No event returned')
        self.assertEqual(event1[DSProps.EVENT_TYPE],
                         event2[DSProps.EVENT_TYPE])
        self.assertEqual(event1[KProps.RESOURCE_NAME],
                         event2[KProps.RESOURCE_NAME])
        self.assertEqual(event1[KProps.RESOURCE_TYPE],
                         event2[KProps.RESOURCE_TYPE])
        self.assertEqual(event1[KProps.PRIORITY],
                         event2[KProps.PRIORITY])



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import werkzeug.http

from http import client as httplib
from keystoneauth1.identity.generic import password
from keystoneauth1 import loading
from keystoneauth1 import session
from keystonemiddleware.auth_token import AuthProtocol
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from webob import exc


LOG = logging.getLogger(__name__)
CFG_GROUP = "service_credentials"


class BasicAndKeystoneAuth(AuthProtocol):

    def __init__(self, application, conf):
        super(BasicAndKeystoneAuth, self).__init__(application, conf)

        self.application = application

        self.oslo_conf = cfg.ConfigOpts()
        self.oslo_conf([],
                       project='vitrage',
                       validate_default_values=True)
        password_option = loading.get_auth_plugin_conf_options('password')
        self.oslo_conf.register_opts(password_option, group=CFG_GROUP)
        self.auth_url = self.oslo_conf.service_credentials.auth_url or ''

    @property
    def reject_auth_headers(self):
        header_val = 'Keystone uri=\'%s\'' % self.auth_url
        return [('WWW-Authenticate', header_val)]

    def process_request(self, req):
        """Process request.

            Evaluate the headers in a request and attempt to authenticate the
            request according to authentication mode.
            If the request comes through /v1/event api path then it can be
            authenticate either with basic auth by providing username and
            password or with keystone authentication.
            If authenticated then additional headers are added to the
            request for use by applications. If not authenticated the request
            will be rejected or marked unauthenticated depending on
            configuration.
        """
        if req.path == '/v1/event':
            basic_auth_info = self._get_basic_authenticator(req)
            if basic_auth_info:
                self._basic_authenticate(basic_auth_info, req)

            else:
                super(BasicAndKeystoneAuth, self).process_request(req)
        else:
            super(BasicAndKeystoneAuth, self).process_request(req)

    def _basic_authenticate(self, auth_info, req):
        try:
            project_domain_id, project_name, user_domain_id = \
                self._get_auth_params()
            auth = password.Password(auth_url=self.auth_url,
                                     username=auth_info.username,
                                     password=auth_info.password,
                                     user_domain_id=user_domain_id,
                                     project_domain_id=project_domain_id,
                                     project_name=project_name)
            sess = session.Session(auth=auth)
            token = sess.get_token()
            project_id = str(auth.get_project_id(sess))
            roles = str(auth.get_auth_ref(sess).role_names[0])
            self._set_req_headers(req, token, project_id, roles)
        except Exception as e:
            to_unicode = encodeutils.exception_to_unicode(e)
            message = 'Authorization exception: %s' % to_unicode
            self._unauthorized(message)

    def _get_auth_params(self):
        user_domain_id = \
            self.oslo_conf.service_credentials.user_domain_id
        project_domain_id = \
            self.oslo_conf.service_credentials.project_domain_id
        project_name = self.oslo_conf.service_credentials.project_name
        return project_domain_id, project_name, user_domain_id

    def _unauthorized(self, message):
        body = {'error': {
            'code': httplib.UNAUTHORIZED,
            'title': httplib.responses.get(httplib.UNAUTHORIZED),
            'message': message,
        }}

        raise exc.HTTPUnauthorized(body=jsonutils.dumps(body),
                                   headers=self.reject_auth_headers,
                                   charset='UTF-8',
                                   content_type='application/json')

    @staticmethod
    def _get_basic_authenticator(req):
        auth = werkzeug.http.parse_authorization_header(
            req.headers.get("Authorization"))
        return auth

    @staticmethod
    def _set_req_headers(req, token, project_id, roles):
        req.headers['X-Auth-Token'] = token
        req.headers['X-Identity-Status'] = 'Confirmed'
        req.headers['X-Project-Id'] = project_id
        req.headers['X-Roles'] = roles


def filter_factory(global_conf, **local_conf):
    """Return a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_filter(app):
        return BasicAndKeystoneAuth(app, conf)

    return auth_filter



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa

from vitrage.storage.sqlalchemy import models


"""Initial migration"

Revision ID: 4e44c9414dff
Revises: None
Create Date: 2019-09-04 15:35:01.086784

"""

# revision identifiers, used by Alembic.
revision = '4e44c9414dff'
down_revision = None


def upgrade():
    try:
        op.create_table(
            'alarms',
            sa.Column('vitrage_id', sa.String(128), primary_key=True),
            sa.Column('start_timestamp', sa.DateTime, nullable=False),
            sa.Column('end_timestamp', sa.DateTime, nullable=False,
                      default=models.DEFAULT_END_TIME),
            sa.Column('name', sa.String(256), nullable=False),
            sa.Column('vitrage_type', sa.String(64), nullable=False),
            sa.Column('vitrage_aggregated_severity', sa.String(64),
                      nullable=False),
            sa.Column('vitrage_operational_severity', sa.String(64),
                      nullable=False),
            sa.Column('project_id', sa.String(64)),
            sa.Column('vitrage_resource_type', sa.String(64)),
            sa.Column('vitrage_resource_id', sa.String(64)),
            sa.Column('vitrage_resource_project_id', sa.String(64)),
            sa.Column('payload', models.JSONEncodedDict),

            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )

        op.create_table(
            'edges',
            sa.Column('source_id', sa.String(128), primary_key=True),
            sa.Column('target_id', sa.String(128), primary_key=True),
            sa.Column('label', sa.String(64), nullable=False),
            sa.Column('start_timestamp', sa.DateTime, nullable=False),
            sa.Column('end_timestamp', sa.DateTime, nullable=False,
                      default=models.DEFAULT_END_TIME),
            sa.Column('payload', models.JSONEncodedDict),

            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            sa.ForeignKeyConstraint(['source_id'], ['alarms.vitrage_id'],
                                    ondelete='CASCADE'),
            sa.ForeignKeyConstraint(['target_id'], ['alarms.vitrage_id'],
                                    ondelete='CASCADE'),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )

        op.create_table(
            'changes',
            sa.Column('id', models.MagicBigInt, primary_key=True,
                      autoincrement=True),
            sa.Column('vitrage_id', sa.String(128), nullable=False),
            sa.Column('timestamp', sa.DateTime, nullable=False),
            sa.Column('severity', sa.String(64), nullable=False),
            sa.Column('payload', models.JSONEncodedDict),

            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            sa.ForeignKeyConstraint(['vitrage_id'], ['alarms.vitrage_id'],
                                    ondelete='CASCADE'),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )

        op.create_table(
            'active_actions',
            sa.Column('action_type', sa.String(128)),
            sa.Column('extra_info', sa.String(128)),
            sa.Column('source_vertex_id', sa.String(128)),
            sa.Column('target_vertex_id', sa.String(128)),
            sa.Column('action_id', sa.String(128), primary_key=True),
            sa.Column('score', sa.SmallInteger()),
            sa.Column('trigger', sa.String(128), primary_key=True),
            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )
        op.create_table(
            'templates',
            sa.Column('id', sa.String(64), primary_key=True, nullable=False),
            sa.Column('status', sa.String(16)),
            sa.Column('status_details', sa.String(128)),
            sa.Column('name', sa.String(128), nullable=False),
            sa.Column('file_content', models.JSONEncodedDict, nullable=False),
            sa.Column("type", sa.String(64), default='standard'),
            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )
        op.create_table(
            'webhooks',
            sa.Column('id', sa.String(128), primary_key=True),
            sa.Column('project_id', sa.String(128), nullable=False),
            sa.Column('is_admin_webhook', sa.Boolean, nullable=False),
            sa.Column('url', sa.String(256), nullable=False),
            sa.Column('headers', sa.String(1024)),
            sa.Column('regex_filter', sa.String(512)),
            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )
        op.create_index(
            'ix_active_action',
            'active_actions',
            [
                'action_type', 'extra_info', 'source_vertex_id',
                'target_vertex_id'
            ]
        )

        op.create_table(
            'events',
            sa.Column("id", sa.BigInteger, primary_key=True, nullable=False,
                      autoincrement=True),
            sa.Column('payload', models.JSONEncodedDict(), nullable=False),
            sa.Column('is_vertex', sa.Boolean, nullable=False),
            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )
        op.create_table(
            'graph_snapshots',
            sa.Column('id', sa.Integer, primary_key=True,
                      nullable=False),
            sa.Column('event_id', sa.BigInteger, nullable=False),
            sa.Column('graph_snapshot', models.CompressedBinary((2 ** 32) - 1),
                      nullable=False),
            sa.Column('created_at', sa.DateTime,
                      default=lambda: timeutils.utcnow()),
            sa.Column('updated_at', sa.DateTime,
                      onupdate=lambda: timeutils.utcnow()),
            mysql_charset='utf8',
            mysql_engine='InnoDB'
        )

        op.create_index(
            'ix_alarms_end_timestamp',
            'alarms',
            [
                'end_timestamp'
            ]
        )

        op.create_index(
            'ix_alarms_project_id',
            'alarms',
            [
                'project_id'
            ]
        )

        op.create_index(
            'ix_alarms_start_timestamp',
            'alarms',
            [
                'start_timestamp'
            ]
        )

        op.create_index(
            'ix_alarms_vitrage_aggregated_severity',
            'alarms',
            [
                'vitrage_aggregated_severity'
            ]
        )

        op.create_index(
            'ix_alarms_vitrage_operational_severity',
            'alarms',
            [
                'vitrage_operational_severity'
            ]
        )

        op.create_index(
            'ix_alarms_vitrage_resource_project_id',
            'alarms',
            [
                'vitrage_resource_project_id'
            ]
        )

        op.create_index(
            'ix_changes_severity',
            'changes',
            [
                'severity'
            ]
        )
        op.create_index(
            'ix_changes_timestamp',
            'changes',
            [
                'timestamp'
            ]
        )

        op.create_index(
            'ix_changes_vitrage_id',
            'changes',
            [
                'vitrage_id'
            ]
        )
    except Exception:
        # TODO(e0ne): figure out more specific exception here to handle a case
        # when migration is applied over Queens release and tables are already
        # exists
        pass



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import threading
import time

from oslo_config import cfg
from oslo_log import log
import oslo_messaging

from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.utils import spawn
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.entity_graph import driver_exec
from vitrage.entity_graph import get_graph_driver

from vitrage.entity_graph.consistency.consistency_enforcer import \
    ConsistencyEnforcer
from vitrage.entity_graph import EVALUATOR_TOPIC
from vitrage.entity_graph.graph_persistency import GraphPersistency
from vitrage.entity_graph.processor.notifier import GraphNotifier
from vitrage.entity_graph.processor.notifier import PersistNotifier
from vitrage.entity_graph.processor.processor import Processor
from vitrage.entity_graph.scheduler import Scheduler
from vitrage.graph.driver.networkx_graph import NXGraph
from vitrage import messaging
from vitrage import storage

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class VitrageGraphInit(object):
    def __init__(self, workers):
        self.graph = get_graph_driver()('Entity Graph')
        self.db = db_connection = storage.get_connection_from_config()
        self.workers = workers
        self.events_coordination = EventsCoordination(self.process_event)
        self.persist = GraphPersistency(db_connection, self.graph)
        self.driver_exec = driver_exec.DriverExec(
            self.events_coordination.handle_multiple_low_priority,
            self.persist, self.graph)
        consistency = ConsistencyEnforcer(
            self.graph,
            self.events_coordination.handle_multiple_high_priority)
        self.scheduler = Scheduler(self.graph,
                                   self.driver_exec,
                                   self.persist,
                                   consistency)

        self.processor = Processor(self.graph)

    def run(self):
        LOG.info('Init Started')
        graph_snapshot = self.persist.query_recent_snapshot()
        if graph_snapshot:
            t = spawn(self.workers.submit_read_db_graph)
            self._restart_from_stored_graph(graph_snapshot)
            t.join()
            self.workers.submit_enable_evaluations()

        else:
            self._start_from_scratch()
            self.workers.submit_read_db_graph()
            self.workers.submit_start_evaluations()
        self._init_finale(immediate_get_all=True if graph_snapshot else False)

    def _restart_from_stored_graph(self, graph_snapshot):
        LOG.info('Main process - loading graph from database snapshot (%sKb)',
                 len(graph_snapshot.graph_snapshot) / 1024)
        NXGraph.read_gpickle(graph_snapshot.graph_snapshot, self.graph)
        self.persist.replay_events(self.graph, graph_snapshot.event_id)
        self._recreate_transformers_id_cache()
        LOG.info("%s vertices loaded", self.graph.num_vertices())
        self.subscribe_presist_notifier()

    def _start_from_scratch(self):
        LOG.info('Starting for the first time')
        LOG.info('Clearing database active_actions')
        self.db.active_actions.delete()
        LOG.info('Disabling previously active alarms')
        self.db.history_facade.disable_alarms_in_history()
        self.subscribe_presist_notifier()
        self.driver_exec.snapshot_get_all()
        LOG.info("%s vertices loaded", self.graph.num_vertices())

    def _init_finale(self, immediate_get_all):
        self._add_graph_subscriptions()
        self.scheduler.start_periodic_tasks(immediate_get_all)
        LOG.info('Init Finished')
        self.events_coordination.start()

    def process_event(self, event):
        if isinstance(event, list):
            for e in event:
                self.processor.process_event(e)
        elif event.get('template_action'):
            self.workers.submit_template_event(event)
            self.workers.submit_evaluators_reload_templates()
        else:
            self.processor.process_event(event)
        self.persist.flush_events()

    def _recreate_transformers_id_cache(self):
        for v in self.graph.get_vertices():
            if not v.get(VProps.VITRAGE_CACHED_ID):
                LOG.warning("Missing vitrage_cached_id in the vertex. "
                            "Vertex is not added to the ID cache %s", v)
            else:
                TransformerBase.key_to_uuid_cache[v[VProps.VITRAGE_CACHED_ID]]\
                    = v.vertex_id

    def _add_graph_subscriptions(self):
        self.graph.subscribe(self.workers.submit_graph_update)
        vitrage_notifier = GraphNotifier()
        if vitrage_notifier.enabled:
            self.graph.subscribe(vitrage_notifier.notify_when_applicable)
            LOG.info('Subscribed vitrage notifier to graph changes')
        self.graph.subscribe(self.persist.persist_event,
                             finalization=True)

    def subscribe_presist_notifier(self):
        self.graph.subscribe(PersistNotifier().notify_when_applicable)


PRIORITY_DELAY = 0.05


class EventsCoordination(object):
    def __init__(self, do_work_func):
        self._lock = threading.Lock()
        self._high_event_finish_time = 0

        def do_work(event):
            try:
                return do_work_func(event)
            except Exception:
                LOG.exception('Got Exception for event %s' % event)

        self._do_work_func = do_work

        self._low_pri_listener = None
        self._high_pri_listener = None

    def start(self):
        self._low_pri_listener = driver_exec.DriversNotificationEndpoint(
            self.handle_multiple_low_priority).init().get_listener()
        self._high_pri_listener = self._init_listener(
            EVALUATOR_TOPIC,
            self._do_high_priority_work)
        LOG.info('Listening on %s', self._high_pri_listener.targets[0].topic)
        LOG.info('Listening on %s', self._low_pri_listener.targets[0].topic)
        self._high_pri_listener.start()
        self._low_pri_listener.start()

    def stop(self):
        self._low_pri_listener.stop()
        self._high_pri_listener.stop()

    def wait(self):
        self._low_pri_listener.wait()
        self._high_pri_listener.wait()

    def _do_high_priority_work(self, event):
        self._lock.acquire()
        self._do_work_func(event)
        self._high_event_finish_time = time.time()
        self._lock.release()

    def _do_low_priority_work(self, event):
        while True:
            self._lock.acquire()
            if (time.time() - self._high_event_finish_time) < PRIORITY_DELAY:
                self._lock.release()
                time.sleep(PRIORITY_DELAY)
            else:
                break
        self._do_work_func(event)
        self._lock.release()

    def handle_multiple_low_priority(self, events):
        index = 0
        if events is None:
            events = []
        for index, e in enumerate(events, 1):
            self._do_low_priority_work(e)
        return index

    def handle_multiple_high_priority(self, events):
        for e in events:
            self._do_high_priority_work(e)

    def _init_listener(self, topic, callback):
        if not topic:
            return
        return messaging.get_notification_listener(
            transport=messaging.get_transport(),
            targets=[oslo_messaging.Target(topic=topic)],
            endpoints=[PushNotificationsEndpoint(callback)])


class PushNotificationsEndpoint(object):
    def __init__(self, process_event_callback):
        self.process_event_callback = process_event_callback

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        try:
            self.process_event_callback(payload)
        except Exception:
            LOG.exception('Failed to process event callback.')



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.nagios.properties import NagiosProperties \
    as NagiosProps
from vitrage.tests import base


class NagiosBaseTest(base.BaseTest):
    def _assert_contains(self, expected_service, services):
        for service in services:
            if service[NagiosProps.RESOURCE_NAME] == \
                    expected_service[NagiosProps.RESOURCE_NAME] and \
                    service[NagiosProps.SERVICE] == \
                    expected_service[NagiosProps.SERVICE]:
                self._assert_expected_service(expected_service, service)
                return

        self.fail("service not found: %(resource_name)s %(service_name)s" %
                  {'resource_name':
                   expected_service[NagiosProps.RESOURCE_NAME],
                   'service_name':
                   expected_service[NagiosProps.SERVICE]})

    def _assert_expected_service(self, expected_service, service):
        for key, value in expected_service.items():
            self.assertEqual(value, service[key], 'wrong value for ' + key)



# Copyright 2019 - Viettel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple

from oslo_log import log

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.datasources.alarm_driver_base import AlarmDriverBase
from vitrage.datasources.kapacitor.config import KapacitorConfig
from vitrage.datasources.kapacitor import KAPACITOR_DATASOURCE
from vitrage.datasources.kapacitor.properties import KapacitorProperties \
    as KProps
from vitrage.datasources.kapacitor.properties import KapacitorState

LOG = log.getLogger(__name__)


class KapacitorDriver(AlarmDriverBase):
    ServiceKey = namedtuple('ServiceKey', ['hostname', 'alarmid'])
    conf_map = None

    def __init__(self):
        super(KapacitorDriver, self).__init__()

        if not KapacitorDriver.conf_map:
            self.conf_map = KapacitorConfig()
        self._client = None

    @staticmethod
    def get_event_types():
        return ['kapacitor.alarm.ok',
                'kapacitor.alarm.info',
                'kapacitor.alarm.warning',
                'kapacitor.alarm.critical']

    def _vitrage_type(self):
        return KAPACITOR_DATASOURCE

    def _alarm_key(self, alarm):
        return self.ServiceKey(hostname=alarm[KProps.RESOURCE_NAME],
                               alarmid=alarm[KProps.ID])

    def _enrich_alarms(self, alarms):
        """Enrich kapacitor alarm using kapacitor configuration file

        Converting Kapacitor host name to Vitrage resource type and name
        It is function of get_all for pulling method
        Not implement yet
        """
        pass

    def enrich_event(self, event, event_type):
        event[DSProps.EVENT_TYPE] = event_type

        kapacitor_host = event[KProps.HOST]
        vitrage_resource = self.conf_map.get_vitrage_resource(kapacitor_host)
        event[KProps.RESOURCE_TYPE] = \
            vitrage_resource[0] if vitrage_resource else None
        event[KProps.RESOURCE_NAME] = \
            vitrage_resource[1] if vitrage_resource else None
        return KapacitorDriver.make_pickleable([event], KAPACITOR_DATASOURCE,
                                               DatasourceAction.UPDATE)[0]

    def _is_erroneous(self, alarm):
        return alarm and alarm[KProps.PRIORITY] != KapacitorState.OK

    def _status_changed(self, new_alarm, old_alarm):
        return new_alarm and old_alarm and \
            not new_alarm[KProps.PRIORITY] == old_alarm[KProps.PRIORITY]

    def _is_valid(self, alarm):
        return alarm[KProps.RESOURCE_TYPE] is not None and \
            alarm[KProps.RESOURCE_NAME] is not None

    def _get_alarms(self):
        """Query all alarm and send to vitrage

        Not implement yet
        """
        return []

    @staticmethod
    def should_delete_outdated_entities():
        return True



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import itertools

from vitrage.common import utils
from vitrage.tests import base
from vitrage.tests.base import IsEmpty


class UtilsTest(base.BaseTest):

    def _assert_set_equal(self, s1, s2, message):
        self.assert_dict_equal(dict.fromkeys(s1, 0),
                               dict.fromkeys(s2, 0),
                               message)

    def test_get_portion(self):
        all_items = list(range(14))
        self._check_portions(all_items, 4)
        self._check_portions(all_items, 3)
        self._check_portions(all_items, 2)
        self._check_portions(all_items, 1)
        self._check_portions(all_items, 51)
        self._check_portions(all_items, 100)

        all_items = [0]
        self._check_portions(all_items, 10)

        all_items = []
        self._check_portions(all_items, 2)

        self._check_portions_bad_params(all_items, 0, 0)
        self._check_portions_bad_params(all_items, -1, 0)
        self._check_portions_bad_params(all_items, 10, 10)

    def _check_portions_bad_params(self, all_items, num, ind):
        exception = None
        try:
            utils.get_portion(all_items, num, ind)
        except Exception as e:
            exception = e
        self.assertIsNotNone(exception, 'get_portion incorrect params')

    def _check_portions(self, all_items, chunks_count):
        chunks = []
        for i in range(chunks_count):
            chunks.append(set(utils.get_portion(all_items, chunks_count, i)))

        union = (a for a in itertools.chain(*chunks))
        self._assert_set_equal(union, set(all_items), 'chunks union differs')
        combinations = itertools.combinations(range(len(chunks)), 2)
        for i, j in combinations:
            self.assertThat(chunks[i].intersection(chunks[j]), IsEmpty(),
                            "Each two chunks should not have "
                            "intersecting items")
        max_size = len(max(chunks, key=lambda x: len(x)))
        min_size = len(min(chunks, key=lambda x: len(x)))
        expected_max_difference = 1 if len(all_items) % len(chunks) else 0
        self.assertEqual(expected_max_difference, max_size - min_size,
                         'chunks sizes should not differ by more than 1')



# Copyright 2015 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from vitrage.tests.mocks.utils import generate_vals
from vitrage.tests.mocks.utils import merge_vals


class BasicEntityModel(object):
    """A representation of the events generated for an entity.

    Events consist of static information that does not change between events,
    and dynamic information that is generated via regex.
    """

    def __init__(self, dynamic_vals, static_vals=None):
        self.static_vals = static_vals
        self.param_specs = dynamic_vals

    @property
    def params(self):
        """Returns a sample event of this entity.

        :return: a sample event of this entity
        :rtype: dict
        """
        dynamic_vals = generate_vals(self.param_specs)
        if self.static_vals:
            return merge_vals(dynamic_vals, self.static_vals)
        else:
            return dynamic_vals
